"""Management of triggers in NER spreadsheets.

Triggers in NER spreadsheets may have unexpected interactions with each other, and
with the scopes in which they are defined.

When NER sheets grow large, we need extra tooling to diagnose these interactions,
so that the user gets the proper feedback on for improving the sheet.
"""

import collections
from textwrap import dedent

from ..core.files import fileOpen
from ..core.helpers import console

from .helpers import toTokens
from .match import occMatch
from .scopes import locInScope


def hasCommon(tokensA, tokensB):
    """Whether one sequence of tokens interferes with another.

    The idea is: we want to determine whether matches for `tokensA` may interfere
    with matches for `tokensB`.

    This happens if `tokensB` is a sublist of `tokensA`, or if an initial segment of
    `tokensB` coincides with a final segment of `tokensA`.

    Or the same with `tokensA` and `tokensB` reversed.

    **Proposition:** *`tokensA` en `tokensB` have something in common (in the above sense)
    if and only if you can make a text where `tokensA` and `tokensB` have overlapping
    matches.*

    $$Proof**:

    (direction `=>`)

    Suppose `tokensA` and `tokensB` have something in common.

    Let `i`, `j` be the start-end position of the (part of) `tokensB` that occur in
    `tokensA`.  Construct a match for the combination of `tokensA` and
    `tokensB` as follows:

    ```
    tokensA[0:i] + tokensA[i:j] + xxx
    ```

    Two cases:

    1.  `tokensB` is fully contained in `tokensA`.
        Then take for `xxx`: `tokensA[j:]`.
        The result is a match for `tokensA` and hence for `tokensB`.
    2.  `tokensB` is only contained in `tokensA` up to index `k`. By definition
        of common, this means that `tokensB[0:k]` is equal to `tokensA[-k:]` and hence
        that `j == len(tokensA)`.
        Then take for `xxx`: `tokensB[k:]`.

        We then have:

        ```
        tokensA + tokensB[k:] =
        tokensA[0:i] + tokensA[i:] + tokensB[k:]
        tokensA[0:i] + tokensA[i:j] + tokensB[k:] =
        ```

        because `j == len(tokensA)` :

        ```
        tokensA[0:i] + tokensA[i:j] + tokensB[k:] =
        tokensA[0:i] + tokensB[0:k] + tokensB[k:] =
        tokensA[0:i] + tokensB
        ```

        So, this text is a match for `tokensA` and for `tokensB`

        (direction `<=`)

        Suppose we have a text `T` with an overlapping match for `tokensA` and
        `tokensB`.

        Suppose `T[i:j]` is a match for `tokensA` and `T[n:m]` is a match for
        `tokensB`, and `T[i:j]` and `T[m:n]` overlap.

        Two cases:

        1.  one match is contained in the other. We consider `T[m:n]` is
            contained in `T[i:j]`. For the reverse case, the argument is the same
            with `tokensA` and `tokensB` interchanged.

            `T[m:n]` is part of a match of `tokensA`, so `T[m:n]` occurs in `tokensA`.
            `T[m:n]` is also a match for `tokensB`, so `tokensB == T[m:n]`, so `tokensB`
            is a part of `tokensA`, hence, by definition: `tokensA` and `tokensB`
            have something in common.
        2.  the two matches have a region in common, but none is contained in the
            other.
            We consider the case where m is between i and j. The case where i is
            between m and n is analogous, with `tokensA` and `tokensB` interchanged.

            Now `T[m:j]` is part of a match for `tokensA` and for `tokensB`.
            Then `T[m:j]` is at the end of `T[i:j]`, so an initial part of
            `tokensB` is a final part of `tokensA`.

    Parameters
    ----------
    tokensA: iterable of string
        first operand
    tokensB: iterable of string
        second operand

    Returns
    -------
    tuple of integer
        We return a result consisting of 3 integers: `ref`, `pos`, `length`

        `ref`: 0 if the two operands are identical; 1 if the first operand
        properly contains the second one or if the second one starts somewhere
        in the first one; -1 otherwise.

        `pos`: the position in the one operand where the other starts.

        `length`: the length of the common part of the two operands.
    """
    nA = len(tokensA)
    nB = len(tokensB)

    if tokensA == tokensB:
        return (0, 0, nA)

    for i in range(nA - 1, -1, -1):
        end = min((nB, nA - i))

        if tokensA[i : i + end] == tokensB[0:end]:
            ref = 1 if i > 0 else 1 if nA >= nB else -1
            return (ref, i, end)

    for i in range(nB - 1, -1, -1):
        end = min((nA, nB - i))

        if tokensB[i : i + end] == tokensA[0:end]:
            ref = -1 if i > 0 else -1 if nB >= nA else 1
            return (ref, i, end)

    return None


def testCommon():
    """Test suite for testing the `hasCommon()` function."""

    tokensA = list("abcd")
    tokensB = list("cdef")
    tokensC = list("defg")
    tokensD = list("bc")
    tokensE = list("ab")
    tokensF = list("cd")
    tokensG = list("a")
    assert hasCommon(tokensA, tokensA) == (0, 0, 4), hasCommon(tokensA, tokensA)
    assert hasCommon(tokensA, tokensB) == (1, 2, 2), hasCommon(tokensA, tokensB)
    assert hasCommon(tokensB, tokensA) == (-1, 2, 2), hasCommon(tokensB, tokensA)
    assert hasCommon(tokensA, tokensC) == (1, 3, 1), hasCommon(tokensA, tokensC)
    assert hasCommon(tokensC, tokensA) == (-1, 3, 1), hasCommon(tokensC, tokensA)
    assert hasCommon(tokensA, tokensD) == (1, 1, 2), hasCommon(tokensA, tokensD)
    assert hasCommon(tokensD, tokensA) == (-1, 1, 2), hasCommon(tokensD, tokensA)
    assert hasCommon(tokensA, tokensE) == (1, 0, 2), hasCommon(tokensA, tokensE)
    assert hasCommon(tokensE, tokensA) == (-1, 0, 2), hasCommon(tokensE, tokensA)
    assert hasCommon(tokensA, tokensF) == (1, 2, 2), hasCommon(tokensA, tokensF)
    assert hasCommon(tokensF, tokensA) == (-1, 2, 2), hasCommon(tokensF, tokensA)
    assert hasCommon(tokensE, tokensG) == (1, 0, 1), hasCommon(tokensE, tokensG)
    assert hasCommon(tokensG, tokensE) == (-1, 0, 1), hasCommon(tokensG, tokensE)

    tokensA = list("abcd")
    tokensB = list("cef")
    tokensC = list("efg")
    tokensD = list("bd")
    tokensE = list("ac")
    tokensF = list("ad")

    assert hasCommon(tokensA, tokensB) is None, hasCommon(tokensA, tokensB)
    assert hasCommon(tokensB, tokensA) is None, hasCommon(tokensB, tokensA)
    assert hasCommon(tokensA, tokensC) is None, hasCommon(tokensA, tokensC)
    assert hasCommon(tokensC, tokensA) is None, hasCommon(tokensC, tokensA)
    assert hasCommon(tokensA, tokensD) is None, hasCommon(tokensA, tokensD)
    assert hasCommon(tokensD, tokensA) is None, hasCommon(tokensD, tokensA)
    assert hasCommon(tokensA, tokensE) is None, hasCommon(tokensA, tokensE)
    assert hasCommon(tokensE, tokensA) is None, hasCommon(tokensE, tokensA)
    assert hasCommon(tokensA, tokensF) is None, hasCommon(tokensA, tokensF)
    assert hasCommon(tokensF, tokensA) is None, hasCommon(tokensF, tokensA)


def makePartition(triggers, myToTokens):
    """Partition a set of triggers into groups of pairwise non-interfering triggers.

    The intention is to explore all triggers that apparently do not have hits.
    We need to look them up in isolation, because then they might have hits on
    their own that are overshadowed by other triggers.

    But searching per trigger is expensive. We want to group triggers together
    that can not interact with each other.
    A hit of one trigger can then never be part of a hit of any other trigger
    in the group.

    "Not being able to interact" is captured by the function `hasCommon()`.

    Parameters
    ----------
    triggers: iterable of string
        The triggers that must be partitioned
    myToTokens: function
        Takes a trigger (string) and produces a sequence of tokens.
        Here you can pass a function that corresponds to how the strings in the corpus
        are divided up into tokens.

    Returns
    -------
    tuple
        The first member is a dict, mapping triggers to sequences of the tokens of which
        they consist;

        the second member is the partition itself, which is a list of parts, where each
        part is a list of triggers that do not have conflicting pairs of triggers.
    """

    triggerTokens = {}
    nTriggerTokens = {}

    for trigger in triggers:
        tokens = myToTokens(trigger)
        triggerTokens[trigger] = tokens
        nTriggerTokens.setdefault(len(tokens), {})[trigger] = tokens

    singleTokenTriggers = list(nTriggerTokens[1]) if 1 in nTriggerTokens else []

    partition = [singleTokenTriggers]

    for n in sorted(nTriggerTokens):
        if n == 1:
            continue
        for triggerA, tokensA in nTriggerTokens[n].items():
            added = False

            for part in partition:
                common = False

                for triggerB in part:
                    tokensB = triggerTokens[triggerB]

                    if hasCommon(tokensA, tokensB):
                        common = True
                        break

                if common:
                    continue

                part.append(triggerA)
                added = True
                break

            if not added:
                partition.append([triggerA])

    return (triggerTokens, partition)


class Triggers:
    def partitionTriggers(self, triggers):
        """Wrapper around the partitioning function.

        This function calls `tf.ner.triggers.makePartition()` with the corpus
        dependent parameter for turning strings into tokens.

        Parameters
        ----------
        triggers: iterable of string
            A sequence of triggers.

        Returns
        -------
        tuple
            The same as what `tf.ner.triggers.makePartition()` returns.
        """
        return makePartition(triggers, self.getToTokensFunc())[1]

    def findOccs(self):
        """Finds the occurrences of multiple triggers.

        This is meant to efficiently list all occurrences of many token
        sequences in the corpus.

        The triggers are in member `instructions`, which must first
        be constructed by reading a number of excel files.

        It adds the member `inventory` to the object, which is a dict
        with subdicts:

        `occurrences`: keyed by tuples (eid, kind), the values are
        the occurrences of that entity in the corpus.
        A single occurrence is represented as a tuple of slots.

        `names`: keyed by tuples (eid, kind) and then path,
        the value is the name of that entity in the context indicated by path.

        """
        if not self.properlySetup:
            return

        settings = self.settings
        spaceEscaped = settings.spaceEscaped

        setData = self.getSetData()
        tokensFromNode = self.tokensFromNode
        getSeqFromNode = self.getSeqFromNode

        buckets = setData.buckets or ()
        sheetData = self.getSheetData()
        caseSensitive = sheetData.caseSensitive
        instructions = sheetData.instructions

        sheetData.inventory = occMatch(
            tokensFromNode,
            getSeqFromNode,
            buckets,
            instructions,
            spaceEscaped,
            caseSensitive=caseSensitive,
        )

    def reportHits(self, silent=None, showNoHits=False):
        """Reports the inventory.

        It diagnoses all triggers by means of `diagnoseTriggers()`.

        Parameters
        ----------
        silent: boolean, optional None
            Whether to be silent. If None, it is taken from the corresponding member
            of the instance.
        showNoHits: boolean, optional False
            Whether to show which triggers do not have hits.
        """
        if not self.properlySetup:
            return

        silent = self.silent if silent is None else silent
        sectionHead = self.sectionHead
        sheetData = self.getSheetData()
        allTriggers = sheetData.allTriggers
        inventory = sheetData.inventory

        setName = self.setName
        annoDir = self.annoDir
        setDir = f"{annoDir}/{setName}"
        reportFile = f"{setDir}/hits.tsv"
        reportTriggerBySlotFile = f"{setDir}/triggerBySlot.tsv"

        hitData = []
        names = set()
        noHits = set()
        triggersBySlot = {}

        for e in sorted(allTriggers):
            (name, eidkind, trigger, scope) = e

            names.add(name)

            entry = (name, trigger, scope)
            section = ""
            hits = ""

            entInfo = inventory.get(eidkind, None)

            if entInfo is None:
                hitData.append(("!E", *entry, "", 0))
                noHits.add(trigger)
                continue

            triggerInfo = entInfo.get(trigger, None)

            if triggerInfo is None:
                hitData.append(("!T", *entry, "", 0))
                noHits.add(trigger)
                continue

            occs = triggerInfo.get(scope, None)

            if occs is None or len(occs) == 0:
                hitData.append(("!P", *entry, "", 0))
                noHits.add(trigger)
                continue

            sectionInfo = collections.Counter()

            for slots in occs:
                for slot in slots:
                    triggersBySlot.setdefault(slot, set()).add(trigger)

                section = sectionHead(slots[0])
                sectionInfo[section] += 1

            for section, hits in sorted(sectionInfo.items()):
                hitData.append(("OK", *entry, section, hits))

        multipleTriggers = {}
        triggerBySlot = {}
        self.triggerBySlot = triggerBySlot

        for slot, triggers in triggersBySlot.items():
            if len(triggers) > 1:
                multipleTriggers[slot] = triggers

            triggerBySlot[slot] = list(triggers)[0]

        if len(multipleTriggers) == 0:
            self.console("No slot is covered by more than one trigger")
        else:
            console(
                f"Slots covered by multiple triggers: {len(multipleTriggers)}",
                error=True,
            )
            for slot, triggers in multipleTriggers.items():
                triggersRep = ", ".join(f"«{trigger}»" for trigger in sorted(triggers))
                self.console(f"{slot:>7}: {triggersRep}", error=True)

        trigWithout = len(noHits)

        if showNoHits and (trigWithout > 0):
            uncovered = 0
            console(
                "Triggers without hits: " f"{trigWithout}x:",
                error=True,
            )

            if len(noHits):
                uncovered = self.diagnoseTriggers(noHits, detail=False)

        with fileOpen(reportFile, "w") as rh:
            rh.write("label\tname\ttrigger\tscope\tsection\thits\n")

            for h in sorted(hitData):
                line = "\t".join(str(c) for c in h)
                rh.write(f"{line}\n")

        with fileOpen(reportTriggerBySlotFile, "w") as rh:
            rh.write("slot\ttrigger\n")

            for slot, trigger in sorted(
                triggerBySlot.items(), key=lambda x: (x[1], x[0])
            ):
                rh.write(f"{slot}\t{trigger}\n")

        nEnt = len(names)
        nTriggers = len(allTriggers)
        nHits = sum(e[-1] for e in hitData)

        msg = (
            f"\t{nEnt} entities targeted with {nHits} occurrences. See {reportFile}"
            if silent
            else dedent(
                f"""
                Entities targeted:          {nEnt:>5}
                Triggers searched for:      {nTriggers:>5}
                Triggers without hits:      {trigWithout:>5}
                 - completely covered:      {trigWithout - uncovered:>5}
                 - missing hits:            {uncovered:>5}
                Triggers with hits:         {nTriggers - trigWithout:>5}
                Total hits:                 {nHits:>5}

                All hits in report file:      {reportFile}
                Triggers by slot in file:     {reportTriggerBySlotFile}
                """
            )
        )
        console(msg)

    def triggerInterference(self, alsoInternal=False, alsoExpected=False):
        """Produce a report of interferences between triggers.

        Triggers interfere if they have matches that intersect, i.e. there is a match
        `m1` of trigger `t1` and a match `m2` of trigger `t2` such that `m1` and
        `m2` intersect (see `tf.ner.triggers.hasCommon()`.

        Triggers may interfere *potentially*: if the triggers have something in
        common they can have overlapping matches. But it does not mean that
        the corpus contains these overlapping matches, i.e. that the triggers conflict
        *actually*.

        We report the *actually* interfering triggers only.

        Triggers within one row are associated to the same entity and work in the same
        row. It is not bad if they are interfering with each other. If there
        are overlapping matches, the trigger that wins still flags the same entity.
        The worst thing is that some of these triggers are superfluous, but there is
        no reason to be picky on superfluous triggers.

        When one trigger is a proper part of another, this is mostly intentional.
        If the longer trigger matches, it wins it from the shorter trigger, unless
        the shorter trigger's match starts before the longer trigger's match.

        We think the user expects the longer trigger to win, but it may surprise him
        if the shorter triggers wins because it starts earlier.

        Parameters
        ----------
        alsoInternal: boolean, optional False
            Also report interference between triggers on the same row.
        alsoExpected: boolean, optional False
            Also report expected interferences.
        """
        if not self.properlySetup:
            return

        setName = self.setName
        annoDir = self.annoDir
        setDir = f"{annoDir}/{setName}"
        reportFile = f"{setDir}/interference.txt"

        app = self.app
        L = app.api.L
        T = app.api.T
        sheetData = self.getSheetData()
        rowMap = sheetData.rowMap
        triggerScopes = sheetData.triggerScopes

        interferences, parts = self.interference(
            rowMap,
            triggerScopes,
            self.getToTokensFunc(),
            self.getSeqFromStr,
            alsoInternal=alsoInternal,
            alsoExpected=alsoExpected,
        )

        messages = []
        witnessed = {}

        nParts = len(parts)
        plural = "" if nParts == 1 else "es"
        self.console(
            f"Looking up {len(interferences)} potential interferences "
            f"in {len(parts)} pass{plural} over the corpus ",
            newline=False,
        )

        for part in parts:
            self.console(".", newline=False)
            inventory = self.findTriggers(part)

            for trigger, data in inventory.items():
                occs = data.get(trigger, {}).get("", [])
                nOccs = len(occs)

                if nOccs:
                    witnessed[trigger] = occs

        self.console("")

        msg = (
            f"{len(witnessed)} potential conflicting trigger pairs with "
            f"{sum(len(x) for x in witnessed.values())} conflicts"
        )
        console(msg)
        messages.append(msg)

        conflicts = {}

        for (
            triggerA,
            triggerB,
            triggerC,
            scopeRepA,
            scopeRepB,
            commonScopes,
        ) in interferences:
            if triggerC not in witnessed:
                continue

            rowA = sorted(set(rowMap[triggerA]))
            rowB = sorted(set(rowMap[triggerB]))
            key = "same row" if rowA == rowB else "different rows"
            conflicts.setdefault(key, []).append(
                (
                    rowA,
                    rowB,
                    triggerA,
                    triggerB,
                    witnessed[triggerC],
                    scopeRepA,
                    scopeRepB,
                    commonScopes,
                )
            )

        for key, confls in conflicts.items():
            newConfls = []

            for (
                rowA,
                rowB,
                triggerA,
                triggerB,
                occs,
                scopeRepA,
                scopeRepB,
                commonScopes,
            ) in confls:
                hits = {}

                for occ in sorted(occs):
                    sectionNode = L.u(occ[0], otype="chunk")[0]
                    heading = tuple(
                        int(x if type(x) is int else x.lstrip("0") or "0")
                        for x in T.sectionFromNode(sectionNode, fillup=True)
                    )

                    if not locInScope(heading, commonScopes):
                        continue

                    heading = app.sectionStrFromNode(sectionNode)
                    hits.setdefault(heading, []).append(occ)

                if len(hits) == 0:
                    continue

                newConfls.append(
                    (rowA, rowB, triggerA, triggerB, hits, scopeRepA, scopeRepB)
                )

            msg = f"{key} ({len(newConfls)} pairs)"
            msg = f"----------\n{msg}\n----------"
            console(msg)
            messages.append(msg)

            for (
                rowA,
                rowB,
                triggerA,
                triggerB,
                hits,
                scopeRepA,
                scopeRepB,
            ) in newConfls:
                rowRepA = ",".join(str(r) for r in rowA)
                rowRepB = ",".join(str(r) for r in rowB)
                msg = (
                    f"{rowRepA:<12} ({scopeRepA:<12}): «{triggerA}»\n"
                    f"{rowRepB:<12} ({scopeRepB:<12}): «{triggerB}»"
                )
                console(msg)
                messages.append(msg)

                diags = []

                i = 0

                for heading, occs in sorted(hits.items()):
                    nOccs = len(occs)

                    if i == 0:
                        diags.append([])

                    diags[-1].append(f"{heading} x {nOccs}")
                    i += 1

                    if i == 5:
                        i = 0

                first = True

                for batch in diags:
                    label = f"{'occurrences':>25}: " if first else (" " * 27)
                    first = False
                    msg = f"{label} {', '.join(batch)}"
                    console(msg)
                    messages.append(msg)

        with fileOpen(reportFile, "w") as fh:
            for msg in messages:
                fh.write(f"{msg}\n")

        console(f"Diagnostic trigger interferences written to {reportFile}")

    def diagnoseTriggers(self, triggers, detail=False):
        """Diagnose triggers individually.

        !!! Caution
            This method will be called by `reportHits()` and should not be called
            on its own, since it expects the instance member `triggerBySlot` which
            is constructed by `reportHits()`

        Here we have a closer look to each trigger individually, not focussing on the
        interaction with other triggers.

        The triggers will be partitioned and each part of the partition will be looked
        up in a separate pass over the corpus.

        Parameters
        ----------
        triggers: iterable of string
            The triggers that must be diagnosed.
        detail: boolean, optional False
            If True, produces more detail per trigger, see `diagnoseTrigger()`

        Returns
        -------
        int
            The number of triggers with uncovered occurrences, see `diagnoseTrigger()`
        """
        if not self.properlySetup:
            return None

        sheetData = self.getSheetData()
        triggerScopes = sheetData.triggerScopes

        parts = self.partitionTriggers(triggers)

        nParts = len(parts)
        plural = "" if nParts == 1 else "es"
        self.console(
            f"Looking up {len(triggers)} triggers "
            f"in {len(parts)} pass{plural} over the corpus ",
            newline=False,
        )

        items = []

        for part in parts:
            self.console(".", newline=False)
            inventory = self.findTriggers(part)

            for trigger, data in inventory.items():
                occs = data.get(trigger, {}).get("", [])
                items.append((trigger, occs))

        uncovered = 0

        for trigger, occs in sorted(
            items,
            key=lambda x: (", ".join(sorted(triggerScopes[x[0]])), x[0].lower()),
        ):
            uncovered += 0 if self.diagnoseTrigger(trigger, occs, detail=detail) else 1

        self.console("")

        return uncovered

    def diagnoseTrigger(self, trigger, occs, detail=False):
        """Diagnoses an individual trigger.

        !!! Caution
            This method will be called by `diagnoseTriggers()` and should not be called
            on its own, since it expects the instance member `triggerBySlot` which
            is constructed by `reportHits()`, which calls `diagnoseTriggers()`.

        All occurrences of the trigger will be checked: is every slot in such
        an occurrence part of a match according to the original spreadsheet?

        If not so, we check whether it is a match of the trigger in question, or of
        another trigger.

        In this way we can detect where the potential but unrealized matches are.

        If there are matches of the trigger in isolation that contain slots that are
        not part of any match of any trigger in the original search, we say that
        the trigger has uncovered occurrences. It will be returned as the result of this
        function whether this is the case.

        These uncovered occurrences are reported as missing hits.

        It can also be the case that the trigger in isolation has matches that
        are completely covered by matches of other triggers in the original search.
        We can also list these, but since these are probably intentional, we suppress
        these cases unless `detail=True` is passed.

        Parameters
        ----------
        trigger: string
            The trigger to be diagnosed.
        occs: iterable of tuple of integer
            The occurrences of this trigger in the whole corpus, irrespective of scope;
        detail: boolean, optional False
            If True, produces complete diagnostics.
            Otherwise, only produces output if there are missed hits.

        Returns
        -------
        boolean
            Whether there are uncovered occurrences.
        """
        if not self.properlySetup:
            return None

        app = self.app
        L = app.api.L
        triggerBySlot = self.triggerBySlot
        sheetData = self.getSheetData()
        triggerScopes = sheetData.triggerScopes

        uncoveredSlots = set()
        coveredBy = {}

        for slots in occs:
            for slot in slots:
                cTrigger = triggerBySlot.get(slot, None)

                if cTrigger is None:
                    uncoveredSlots.add(slot)
                else:
                    coveredBy.setdefault(cTrigger, set()).add(slot)

        uncoveredOccs = {}

        nUncoveredSlots = len(uncoveredSlots)
        ok = nUncoveredSlots == 0

        if nUncoveredSlots:
            for slot in sorted(uncoveredSlots):
                heading = app.sectionStrFromNode(L.u(slot, otype="chunk")[0])
                occ = uncoveredOccs.setdefault(heading, [[]])

                if len(occ[-1]) == 0 or occ[-1][-1] + 1 == slot:
                    occ[-1].append(slot)
                else:
                    occ.append([slot])

        nMissedHits = 0
        uncoveredOccsDiag = []
        uncoveredOccsDiagCompact = []

        for heading, occs in uncoveredOccs.items():
            nOccs = len(occs)
            nMissedHits += nOccs
            uncoveredOccsDiag.append(f"\t\t{heading}: {nOccs} x")
            uncoveredOccsDiagCompact.append(f"{heading} x {nOccs}")

        uncoveredOccsDiag[0:0] = [f"\tuncovered: {nMissedHits} x"]

        coveredOccsDiag = []

        for cTrigger in sorted(coveredBy, key=lambda x: x.lower()):
            thisCoveredOccsDiag = []
            coveredSlots = sorted(coveredBy[cTrigger])

            coveredHits = {}

            for slot in sorted(coveredSlots):
                heading = app.sectionStrFromNode(L.u(slot, otype="chunk")[0])
                occ = coveredHits.setdefault(heading, [[]])

                if len(occ[-1]) == 0 or occ[-1][-1] + 1 == slot:
                    occ[-1].append(slot)
                else:
                    occ.append([slot])

            nCoveredHits = 0

            for heading, occs in coveredHits.items():
                nOccs = len(occs)
                nCoveredHits += nOccs
                thisCoveredOccsDiag.append(f"\t\t{heading}: {nOccs} x")

            thisCoveredOccsDiag[0:0] = [f"\tcovered by: {cTrigger}: {nCoveredHits} x"]

            coveredOccsDiag.extend(thisCoveredOccsDiag)

        scopeRep = f"({', '.join(sorted(triggerScopes[trigger]))})"

        if detail:
            console(f"{trigger} {scopeRep}:")

            for line in uncoveredOccsDiag:
                console(line)

            for line in coveredOccsDiag:
                console(line)

        else:
            if nUncoveredSlots == 0:
                if False:
                    console(f"{trigger} {scopeRep}: covered by other triggers")
            else:
                missedHits = []

                i = 0
                for occRep in uncoveredOccsDiagCompact:
                    if i == 0:
                        missedHits.append([occRep])
                    else:
                        missedHits[-1].append(occRep)

                    i += 1
                    if i == 5:
                        i = 0

                missedHitsRep = ", ".join(missedHits[0])
                console(f"{trigger:<40} {scopeRep:<12}: {missedHitsRep}", error=True)

                for m in missedHits[1:]:
                    missedHitsRep = ", ".join(m)
                    console(f"{' ' * 55}{missedHitsRep}", error=True)

        return ok

    def interference(
        self,
        rowMap,
        triggerScopes,
        myToTokens,
        getSeqFromStr,
        alsoInternal=False,
        alsoExpected=False,
    ):
        """Workhorse to calculate interference of triggers.

        This function is used by method `triggerInterference()` for calculating the
        *potential* interferences within a set of triggers, irrespective whether
        the corpus contains instances of these interferences.

        When two triggers interfere, you can build a combined trigger out of them
        whose matches are exactly the points of interference. We call this the
        *combined* trigger of the interference.

        Parameters
        ----------
        rowMap: dict
            mapping from triggers to the rows where they occur in the spreadsheet.
            We take the set of triggers from here, so that we can issue proper
            diagnostic information later, i.e. we can mentoin the row in the spreadsheet
            where the offending triggers are.
        triggerScopes: dict
            mapping from triggers to the scopes for which they are defined.
        myToTokens: function
            corpus dependent function for parsing a trigger, which is a string,
            to tokens, in the same way as the text of the corpus is parsed into tokens.
        getSeqfromStr: function
            corpus dependent function that can translate a section heading into a
            tuple of integers that represents the same heading in the "legal" numbering
            system.
        alsoInternal: boolean, optional False
            Whether to report interfering triggers even if they belong to the
            same entity
        alsoExpected: boolean, optional False
            Whether to report interfering triggers even if one is a proper
            initial part of the other.

        Returns
        -------
        tuple
            There are two parts:

            *interferences*: a list of interfering pairs of triggers, where each
            interference has these components:

            *   *triggerA* the first trigger in the pair;
            *   *triggerB* the second trigger in the pair;
            *   *triggerC* the combined trigger of the interference;
            *   *scopeRepA* the scope of the first trigger;
            *   *scopeRepB* the scope of the second trigger;
            *   *commonScopes* the intersection of the scopes of both triggers.

            *parts*: a partition of the combined triggers of all interferences so that
            within each part none of these interfere.
        """
        if not self.properlySetup:
            return (None, None)

        triggers = list(rowMap)

        triggerTokens, parts = makePartition(triggers, myToTokens)

        nParts = len(parts)

        interferences = []

        intersections = {}

        for i, part in enumerate(parts):
            if i == nParts - 1:
                break

            for otherPart in parts[i + 1 : nParts]:
                for triggerA in part:
                    for triggerB in otherPart:
                        tokensA = triggerTokens[triggerA]
                        tokensB = triggerTokens[triggerB]

                        if not alsoInternal:
                            rowsA = set(rowMap[triggerA])
                            rowsB = set(rowMap[triggerB])
                            if rowsA == rowsB:
                                continue

                        scopesA = ",".join(sorted(triggerScopes[triggerA]))
                        scopesB = ",".join(sorted(triggerScopes[triggerB]))
                        commonScopes = intersections.get((triggerA, triggerB), None)

                        if commonScopes is None:
                            commonScopes = self.intersectScopes(scopesA, scopesB)
                            intersections[(triggerA, triggerB)] = commonScopes

                        if len(commonScopes) == 0:
                            continue

                        common = hasCommon(tokensA, tokensB)

                        if common is None:
                            continue

                        ref, pos, length = common
                        nTokensA = len(tokensA)
                        nTokensB = len(tokensB)

                        nTokensLatter = nTokensB if ref == 1 else nTokensA

                        expected = length == nTokensLatter

                        if expected and not alsoExpected:
                            continue

                        if ref == 1:
                            nB = len(tokensB)
                            union = tokensA

                            if length < nB:
                                union += tokensB[length:]
                        else:
                            nA = len(tokensA)
                            union = tokensB

                            if length < nA:
                                union += tokensA[length:]

                        interferences.append(
                            (
                                triggerA,
                                triggerB,
                                " ".join(union),
                                scopesA,
                                scopesB,
                                commonScopes,
                            )
                        )

        parts = makePartition([x[2] for x in interferences], myToTokens)[1]

        return interferences, parts

    def findTriggers(self, triggers):
        """Looks up occurrences of multiple triggers efficiently.

        This is a lot like `findOccs()`, but where as `findOccs()` finds
        its search instructions in the sheet data and stores its search results in the
        sheet data, this function makes its own instructions and returns the search
        results.

        We use this function when we need to investigate what triggers do in isolation
        and without scope restrictions.

        So, the instructions generated by this functions are derived from the
        spreadsheet, but ignore all scope restrictions.

        Parameters
        ----------
        triggers: iterable of string
            These are the triggers to be looked up. Typically they come from one
            part of a partition of triggers, but not necessarily so.

        Returns
        -------
        dict
            The inventory of occurrences of the triggers, as returned by
            `tf.ner.match.occMatch()`.
        """
        if not self.properlySetup:
            return {}

        settings = self.settings
        spaceEscaped = settings.spaceEscaped

        setData = self.getSetData()
        tokensFromNode = self.tokensFromNode
        getSeqFromNode = self.getSeqFromNode

        buckets = setData.buckets or ()
        sheetData = self.getSheetData()
        caseSensitive = sheetData.caseSensitive

        idMap = {trigger: trigger for trigger in triggers}
        tMap = {trigger: "" for trigger in triggers}
        tPos = {}
        triggerSet = set()
        instructions = {(): dict(tPos=tPos, tMap=tMap, idMap=idMap)}

        for trigger in triggers:
            triggerT = toTokens(
                trigger, spaceEscaped=spaceEscaped, caseSensitive=caseSensitive
            )
            triggerSet.add(triggerT)

        for triggerT in triggerSet:
            for i, token in enumerate(triggerT):
                tPos.setdefault(i, {}).setdefault(token, set()).add(triggerT)

        inventory = occMatch(
            tokensFromNode,
            getSeqFromNode,
            buckets,
            instructions,
            spaceEscaped,
            caseSensitive=caseSensitive,
        )

        return inventory

    def findTrigger(self, trigger, show=True):
        """Looks up occurrences of a single triggers.

        This is a lot like `findTriggers()`, but where as `findTriggers()` looks up
        multiple triggers, this one looks up a single trigger.

        Like `findTriggers()` the search is done without scope restrictions.

        We use this function when we need to investigate what a single trigger
        does in isolation and without scope restrictions.

        So, the instructions generated by this functions are derived from the
        spreadsheet, but ignore all scope restrictions.

        Parameters
        ----------
        trigger: string
            This is the trigger to be looked up.
        show: boolean, optional True
            If True, it shows the result on the console, otherwise it returns
            the result.

        Returns
        -------
        list or void
            The list of occurrences of the trigger, provided `show=False` is passed,
            else there is no function result.
        """
        if not self.properlySetup:
            return []

        app = self.app
        L = app.api.L

        settings = self.settings
        spaceEscaped = settings.spaceEscaped

        setData = self.getSetData()
        tokensFromNode = self.tokensFromNode
        getSeqFromNode = self.getSeqFromNode

        buckets = setData.buckets or ()
        sheetData = self.getSheetData()
        caseSensitive = sheetData.caseSensitive

        triggerT = toTokens(
            trigger, spaceEscaped=spaceEscaped, caseSensitive=caseSensitive
        )
        idMap = {trigger: trigger}
        tMap = {trigger: ""}
        tPos = {}

        for i, token in enumerate(triggerT):
            tPos.setdefault(i, {}).setdefault(token, set()).add(triggerT)

        instructions = {(): dict(tPos=tPos, tMap=tMap, idMap=idMap)}

        inventory = occMatch(
            tokensFromNode,
            getSeqFromNode,
            buckets,
            instructions,
            spaceEscaped,
            caseSensitive=caseSensitive,
        )

        occs = inventory.get(trigger, {}).get(trigger, {}).get("", [])

        if show:
            nOccs = len(occs)
            plural = "" if nOccs == 1 else "s"
            app.dm(f"**{nOccs} occurrence{plural}**\n")

            if nOccs:
                headings = set()
                highlights = set()

                for occ in occs:
                    headings.add(L.u(occ[0], otype="chunk")[0])

                    for slot in occ:
                        highlights.add(slot)

                for hd in sorted(headings):
                    app.plain(hd, highlights=highlights)
        else:
            return occs



"""Detect name variants in the corpus.

We provide a class can detect variants of the triggers in a NER spreadsheet,
that is variants as they occur in the corpus.

One way (and the way supported here) of obtaining the variants is
by the tool
[analiticcl](https://github.com/proycon/analiticcl) by Martin
Reynaert and Maarten van Gompel.
"""

import collections
import re

# from analiticcl import VariantModel, Weights, SearchParameters

from ..capable import CheckImport
from ..core.files import initTree, fileExists, readJson, writeJson, fileOpen
from ..core.helpers import console
from ..convert.recorder import Recorder


HTML_PRE = """<html>
<head>
    <meta charset="utf-8"/>
    <style>
«css»
    </style>
</head>
<body>
"""

HTML_POST = """
</body>
</html>
"""


class Detect:
    def __init__(self, NE, sheet=None):
        """Detect spelling variants.

        In order to use this class you may have to install analiticcl. First
        you have to make sure you have the programming language Rust operational on
        your machine, and then you have to install analiticcl as a Rust program,
        and finally you have to do

        ```
        pip install analiticcl
        ```

        See the
        [python bindings](https://github.com/proycon/analiticcl/tree/master/bindings/python)
        for detailed instructions.

        If you have an instance of the `tf.ner.ner.NER` class, e.g. from an earlier
        call like this

        ```
        NE = A.makeNer()
        ```

        and if you have pointed this `NE` to a NER spreadsheet, e.g. by

        ```
        NE.setTask(".persons")
        ```

        then you can get an instance of this class by saying

        ```
        D = NE.variantDetection()
        ```

        There are methods to produce a plain text of the corpus, a lexicon based on the
        triggers in the spreadsheet, and then to search the plain text for variants
        of the lexicon by passing some well-chosen parameters to analiticcl.

        After the search, which may take long, the raw results are cached, and then
        filtered, with the help of an optional exceptions file.
        If you change the filtering parameters, you do not have to rerun the expensive
        search by analiticcl.

        The variants found can then be merged with the original triggers, and saved as
        a new spreadsheet, next to the original one.

        The settings for analiticcl are ultimately given in the config.yaml in the
        `ner` directory of the corpus, where the other settings for the NER process
        are also given. See `tf.ner.settings`.

        Parameters
        ----------
        NE: object
            A `tf.ner.ner.NER` instance.
        sheet: string, optional None
            The name of a NER sheet that serves as input of the variant detection
            process. If not passed, it is assumed that the `NE` instances is
            already switched to a NER sheet before setting up this object.
            However, in any case, we switch again to the sheet in question to make
            sure we do so in case sesnsitive mode (even if we have used the sheet
            in case insensitive mode for the lookup).
        """
        CI = CheckImport("analiticcl")

        if CI.importOK(hint=True):
            an = CI.importGet()
            self.VariantModel = an.VariantModel
            self.Weights = an.Weights
            self.SearchParameters = an.SearchParameters
        else:
            self.properlySetup = False
            return

        self.properlySetup = True

        self.NE = NE

        if sheet is None:
            sheet = NE.sheetName
            NE.setSheet(sheet, caseSensitive=True, force=True, forceSilent=True)
        else:
            NE.setSheet(sheet, caseSensitive=True, force=True)

        if not NE.setIsX:
            console(
                (
                    "Setting up this instance requires having a current "
                    f"NER spreadsheet selected. '{sheet}' is not a spreadsheet."
                ),
                error=True,
            )
            self.properlySetup = False
            return

        self.sheet = sheet
        self.sheetData = NE.getSheetData()

        app = NE.app
        self.app = app

        settings = NE.settings.variants
        self.settings = settings

        workDir = f"{app.context.localDir}/{app.context.extraData}/analyticcl"
        self.workDir = workDir
        initTree(workDir, fresh=False)

        NE.setSheet(sheet, caseSensitive=True, force=True)
        sheetData = NE.getSheetData()

        NE.console("Overview of names by length:")
        triggers = set(sheetData.rowMap)

        lengths = collections.defaultdict(list)

        for trigger in triggers:
            lengths[len(trigger.split())].append(trigger)

        for n, trigs in sorted(lengths.items(), key=lambda x: -x[0]):
            examples = "\n      ".join(sorted(trigs, key=lambda x: x.lower())[0:5])
            NE.console(f"  {n} tokens: {len(trigs):>3} names e.g.:\n      {examples}")

    def prepare(self):
        """Prepare the data for the search of spelling variants.

        We construct an alphabet and a plain text out of the corpus,
        and we construct a lexicon from the triggers in the current spreadsheet.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        self.makeAlphabet()
        self.makeText()
        self.makeLexicon()
        self.setupAnaliticcl()

    def search(self, start=None, end=None, force=0):
        """Search for spelling variants in the corpus.

        We search part of the corpus or the whole corpus for spelling variants.
        The process has two stages:

        1.  a run of analaticcl
        2. filtering of the results

        The run of analiticcl is expensive, more than 10 minutes on the
        [Suriano corpus](https://gitlab.huc.knaw.nl/suriano/letters).

        The results of stage 1 will be cached. For every choice of search parameters
        and boundary points in the corpus there is a separate cache.

        A number of analiticcl-specific parameters will influence the search.
        They can be tweaked in the config file of the NER module, under the variants
        section.

        Parameters
        ----------
        start: integer, optional None
            The place in the corpus where the search has to start; it is the offset
            in the plain text. If `None`, start from the beginning of the corpus.
        end: integer, optional None
            The place in the corpus where the search must end; it is the offset
            in the plain text. If `None`, the search will be performed till the end
            of the corpus.
        force: integer, optional 0
            Valid values are `0`, `1`, `2`.
            Meaning: `0`: use the cached result, if it is available.
            `1`: use the cached result for stage 1, if available, but perform the
            filtering (again).
            `2`: do not use the cache, but compute everything again.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        SearchParameters = self.SearchParameters
        ansettings = self.settings.analiticcl
        searchParams = ansettings.searchParams
        suoffsets = searchParams.unicodeoffsets
        smaxngram = searchParams.max_ngram
        sfreqweight = searchParams.freq_weight
        scoring = ansettings.scoring
        sthreshold = scoring.threshold

        NE = self.NE
        app = self.app
        model = self.model
        rec = self.rec
        textComplete = self.textComplete
        workDir = self.workDir
        lexiconOccs = self.lexiconOccs

        text = textComplete[start:end]
        nText = len(text)

        offset = 0 if start is None else nText + start if start < 0 else start

        NE.console(f"{nText:>8} text  length")
        NE.console(f"{offset:>8} offset in complete text")

        slug = (
            f"{suoffsets}-{smaxngram}-{sfreqweight}-{sthreshold}-"
            f"{suoffsets}-{smaxngram}-{sfreqweight}-{sthreshold}"
        )
        matchesFile = f"{workDir}/matches-{start}-{end}-settings={slug}.json"
        matchesPosFile = f"{workDir}/matchespos-{start}-{end}-settings={slug}.json"
        rawMatchesFile = f"{workDir}/rawmatches-{start}-{end}-settings={slug}.json"

        app.indent(reset=True)

        if force == 2 or not fileExists(rawMatchesFile):
            app.info("Compute variants of the lexicon words ...")
            rawMatches = model.find_all_matches(
                text,
                SearchParameters(
                    unicodeoffsets=suoffsets,
                    max_ngram=smaxngram,
                    freq_weight=sfreqweight,
                    score_threshold=sthreshold,
                ),
            )
            writeJson(rawMatches, asFile=rawMatchesFile)
        else:
            app.info("Read previously computed variants of the lexicon words ...")
            rawMatches = readJson(asFile=rawMatchesFile, plain=False)

        app.info(f"{len(rawMatches):>8} raw   matches")

        if force == 1 or not fileExists(matchesFile) or not fileExists(matchesPosFile):
            app.info("Filter variants of the lexicon words ...")
            positions = rec.positions(simple=True)

            matches = {}
            matchPositions = collections.defaultdict(list)

            for match in rawMatches:
                text = match["input"].replace("\n", " ")
                textL = text.lower()

                if text in lexiconOccs:
                    continue

                candidates = match["variants"]

                if len(candidates) == 0:
                    continue

                candidates = {
                    cand["text"]: s
                    for cand in candidates
                    if (s := cand["score"]) >= sthreshold
                }

                if len(candidates) == 0:
                    continue

                textRemove = set()

                for cand in candidates:
                    candL = cand.lower()
                    if candL == textL:
                        textRemove.add(cand)

                for cand in textRemove:
                    del candidates[cand]

                if len(candidates) == 0:
                    continue

                # if the match ends with 's we remove the part without it from the
                # candidates

                if text.endswith("'s"):
                    head = text.removesuffix("'s")
                    if head in candidates:
                        del candidates[head]

                        if len(candidates) == 0:
                            continue

                # we have another need to filter: if the text of a match is one short
                # word longer than a candidate we remove that candidate
                # provided the extra word is lower case and has at most 3 letters
                # this is to prevent cases like
                # «Adam Schivelbergh» versus «Adam Schivelbergh di»
                #
                # We do this also when the extra word is at the start, like
                # «di monsignor Mangot» versus «monsignor Mangot»

                parts = text.split()

                if len(parts) > 0:
                    (head, tail) = (parts[0:-1], parts[-1])

                    # if len(tail) <= 3 and tail.islower():
                    if len(tail) <= 3:
                        head = " ".join(head)

                        if head in candidates:
                            del candidates[head]

                            if len(candidates) == 0:
                                continue

                    (head, tail) = (parts[0], parts[1:])

                    # if len(head) <= 3 and head.islower():
                    if len(head) <= 3:
                        tail = " ".join(tail)

                        if tail in candidates:
                            del candidates[tail]

                            if len(candidates) == 0:
                                continue

                position = match["offset"]
                start = position["begin"]
                end = position["end"]
                nodes = sorted(
                    {positions[i] for i in range(offset + start, offset + end)}
                )

                matches[text] = candidates
                matchPositions[text].append(nodes)

            writeJson(matches, asFile=matchesFile)
            writeJson(matchPositions, asFile=matchesPosFile)
        else:
            app.info("Read previously filtered variants of the lexicon words ...")
            matches = readJson(asFile=matchesFile, plain=False)
            matchPositions = readJson(asFile=matchesPosFile, plain=False)

        app.info(f"{len(matches):>8} filtered matches")

        self.matches = matches
        self.matchPositions = matchPositions
        console(f"{len(matches)} variants found")

    def listVariants(self, start=None, end=None):
        """List the search results to the console.

        Show (part of) the variants found on the console as a plain text table.

        This content will also be written to the file `variants.txt` in the
        work directory.

        Parameters
        ----------
        start: integer, optional None
            The sequence number of the first result to show.
            If `None`, start with the first result.
        end: integer, optional None
            The sequence number of the last result to show.
            If `None`, continue to the last result.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        workDir = self.workDir
        matches = self.matches

        lines = []

        head = ("variant", "score", "candidate")
        dash = f"{'-' * 4} | {'-' * 25} | {'-' * 5} | {'-' * 25}"
        console(f"{'i':>4} | {head[0]:<25} | {head[1]} | {head[2]}")
        console(f"{dash}")
        startN = start or 0

        for text, candidates in sorted(matches.items(), key=lambda x: x[0].lower()):
            for cand, score in sorted(candidates.items(), key=lambda x: x[0].lower()):
                lines.append((text, score, cand))

        for i, (text, score, cand) in enumerate(lines[start:end]):
            console(f"{i + startN:>4} | {text:<25} |  {score:4.2f} | {cand}")

        console(f"{dash}")

        file = f"{workDir}/variants.tsv"

        with fileOpen(file, "w") as fh:
            headStr = "\t".join(head)
            fh.write(f"{headStr}\n")
            for text, score, cand in lines:
                fh.write(f"{text}\t{score:4.2f}\t{cand}\n")

        console(f"{len(matches)} variants found and written to {file}")

    def mergeTriggers(self, level=1):
        """Merge spelling variants of triggers into a NER sheet.

        When we have found spelling variants of triggers, we want to include
        them in the entity lookup. This function places the variants in the same
        cells as the triggers they are variants of. However, it will not
        overwrite the original spreadsheet, but create a new, enriched spreadsheet.

        We collect the necessary information as follows:

        *   `matches: dict`:
            The spelling variants are keys, and their values are again dicts, keyed
            by the words in the triggers that come closest, and valued by a measure
            of the proximity.
            It is assumed that all of these variants are good variants, in that the
            scores are always above a certain threshold, e.g. 0.8 .
        *   `mergedFile: string`:
            The path of the new spreadsheet with the merged triggers: it will sit next
            to the original spreadsheet, but with an extension such as `-merged` added
            to its name (this can be configured in the NER config file near the corpus).
        *   `exclusionFile: string, optional None`
            The path to an optional file with exclusions, one per line.
            Variants that occur in the exclusion list will not be merged in.
            The file sits next to the original spreadsheet, but with an extension such
            as `-notmerged` (this is configurable) and file extension `.txt`.

        This function will also produce several reports:

        Parameters
        ----------
        level: integer, optional 1
            Only relevant for reporting the new variants. Occurrences of the
            new variants are counted by section. This parameter specifies the
            level of those sections. It should be 1, 2 or 3.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        NE = self.NE
        trigI = NE.trigI
        commentI = NE.commentI
        sheetDir = NE.sheetDir
        sheetName = NE.sheetName
        settings = self.settings
        mergedExtension = settings.mergedExtension
        notMergedExtension = settings.notMergedExtension
        mergedFile = f"{sheetDir}/{sheetName}{mergedExtension}.xlsx"
        mergedReportFile = f"{sheetDir}/{sheetName}{mergedExtension}-report.tsv"
        exclusionFile = f"{sheetDir}/{sheetName}{notMergedExtension}.txt"

        if not fileExists(exclusionFile):
            NE.console(f"File with excluded variants not found: {exclusionFile}")
            exclusionFile = None

        matches = self.matches
        matchPositions = self.matchPositions
        sectionHead = NE.sectionHead

        noVariant = set()

        if exclusionFile is not None and fileExists(exclusionFile):
            with fileOpen(exclusionFile) as fh:
                for line in fh:
                    noVariant.add(line.strip())

            nNoVariant = len(noVariant)
            pl = "" if nNoVariant == 1 else "s"
            NE.console(f"{nNoVariant} excluded variant{pl} found in {exclusionFile}")

        mapping = {}
        excluded = 0

        for text, candidates in matches.items():
            if text in noVariant:
                excluded += 1
                continue
            for cand in candidates:
                mapping.setdefault(cand, set()).add(text)

        ple = "" if excluded == 1 else "s"
        NE.console(f"{excluded} variant{ple} excluded as trigger")

        rows = NE.readSheetData()

        nAdded = 0
        totAdded = 0

        variantsAdded = {}

        for r, row in enumerate(rows):
            if r == 0 or r == 1 or row[commentI].startswith("#"):
                continue

            rn = r + 1
            triggers = set(row[trigI])
            nPrev = len(triggers)

            newTriggers = []

            for trigger in triggers:
                newTriggers.append(trigger)

                for variant in mapping.get(trigger, []):
                    newTriggers.append(variant)
                    variantsAdded.setdefault(rn, []).append((variant, trigger))

            newTriggers = sorted(set(newTriggers))
            row[trigI] = newTriggers
            nPost = len(newTriggers)

            nDiff = nPost - nPrev

            if nDiff != 0:
                nAdded += 1
                totAdded += nDiff

        lines = [("row", "trigger", "variant", "occurences")]

        for rn in sorted(variantsAdded):
            for variant, trigger in variantsAdded[rn]:
                sectionInfo = collections.Counter()

                for occ in matchPositions[variant]:
                    slot = occ[0]

                    section = sectionHead(slot, level=level)
                    sectionInfo[section] += 1

                hitData = [
                    f"{section}x{hits}" for section, hits in sorted(sectionInfo.items())
                ]
                for hits in hitData:
                    lines.append((rn, trigger, variant, hits))

        with fileOpen(mergedReportFile, "w") as fh:
            nLines = len(lines)

            for i, line in enumerate(lines):
                if i < 10 or i > nLines - 10:
                    (row, trigger, variant, hits) = line
                    NE.console(f"{row:<4} {trigger:<40} ~> {variant:<40} = {hits}")

                lineStr = "\t".join(str(x) for x in line)
                fh.write(f"{lineStr}\n")

        pls = "" if nAdded == 1 else "s"
        plt = "" if totAdded == 1 else "s"
        NE.console(f"{nAdded} triggerset{pls} expanded with {totAdded} trigger{plt}")
        NE.console(f"Wrote merge report to file {mergedReportFile}")

        NE.writeSheetData(rows, asFile=mergedFile)
        NE.console(f"Wrote merged triggers to sheet {mergedFile}")

    def showResults(self, start=None, end=None):
        """Show the search results to the console.

        Show (part of) the variants found on the console with additional context.

        Parameters
        ----------
        start: integer, optional None
            The sequence number of the first result to show.
            If `None`, start with the first result.
        end: integer, optional None
            The sequence number of the last result to show.
            If `None`, continue to the last result.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        app = self.app
        F = app.api.F
        matches = self.matches
        matchPositions = self.matchPositions

        i = 0

        for text, candidates in sorted(matches.items())[start:end]:
            i += 1
            nCand = len(candidates)
            pl = "" if nCand == 1 else "s"
            console(f"{i:>4} Variant «{text}» of {nCand} candidate{pl}")
            console("  Occurrences:")
            occs = matchPositions[text]

            for nodes in occs:
                sectionStart = app.sectionStrFromNode(nodes[0])
                sectionEnd = app.sectionStrFromNode(nodes[-1])
                section = (
                    sectionStart
                    if sectionStart == sectionEnd
                    else f"{sectionStart} - {sectionEnd}"
                )
                preStart = max((nodes[0] - 10, 1))
                preEnd = nodes[0]
                postStart = nodes[-1] + 1
                postEnd = min((nodes[-1] + 10, F.otype.maxSlot + 1))
                preText = "".join(
                    f"{F.str.v(n)}{F.after.v(n)}" for n in range(preStart, preEnd)
                )
                inText = "".join(f"{F.str.v(n)}{F.after.v(n)}" for n in nodes)
                postText = "".join(
                    f"{F.str.v(n)}{F.after.v(n)}" for n in range(postStart, postEnd)
                )
                context = f"{section}: {preText}«{inText}»{postText}".replace("\n", " ")
                console(f"    {context}")

            console("  Candidates with score:")

            for cand, score in sorted(candidates.items(), key=lambda x: (-x[1], x[0])):
                console(f"\t{score:4.2f} {cand}")

            console("-----")

    def displayResults(self, start=None, end=None, asFile=None):
        """Display the results as HTML files.

        This content will also be written to the files under the subdirectory
        `extra` in the work directory.

        Parameters
        ----------
        start: integer, optional None
            The sequence number of the first result to show.
            If `None`, start with the first result.
        end: integer, optional None
            The sequence number of the last result to show.
            If `None`, continue to the last result.
        asFile: string, optional None
            If None, the results will be displayed as HTML on the console.
            Otherwise, the results will be written down as a set of HTML files,
            whose names start with this string.
        """

        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        app = self.app
        L = app.api.L
        matches = self.matches
        matchPositions = self.matchPositions
        lexiconOccs = self.lexiconOccs
        workDir = self.workDir

        if asFile is not None:
            content = []

            htmlStart = HTML_PRE.replace("«css»", app.context.css)
            htmlEnd = HTML_POST

            content.append([htmlStart])
            empty = True

            app.indent(reset=True)
            app.info("Gathering information on extra triggers ...")

        i = 0
        s = 0

        for varText, candidates in sorted(
            matches.items(), key=lambda x: (-len(x[1]), x[0])
        )[start:end]:
            i += 1

            # make a list of where the candidates are and include the score

            varOccs = matchPositions[varText]
            nVar = len(varOccs)

            candOccs = []
            candRep1 = "`" + "` or `".join(candidates) + "`"
            candRep2 = "<code>" + "</code> or <code>".join(candidates) + "</code>"

            for cand, score in candidates.items():
                myOccs = lexiconOccs[cand]

                for occ in myOccs:
                    candOccs.append((occ[0], score, cand, occ))

            # use this list later to find the nearest/best variant

            if asFile is None:
                app.dm(
                    f"# {i}: {nVar} x variant `{varText}` on "
                    f"candidate {candRep1}\n\n"
                )
            else:
                content[-1].append(
                    f"<h1>{i}: {nVar} x variant <code>{varText}</code> on "
                    f"candidate {candRep2}</h1>"
                )
                empty = False

            sections = set()
            highlights = {}
            bestCand = None

            for candOcc in candOccs:
                if bestCand is None or bestCand[1] < candOcc[1]:
                    bestCand = candOcc

            for n in bestCand[3]:
                highlights[n] = "lightgreen"

            section = L.u(bestCand[0], otype="chunk")[0]
            sections.add(section)

            for varNodes in varOccs:
                highlights |= {n: "goldenrod" for n in varNodes}

                nFirst = varNodes[0]
                section = L.u(nFirst, otype="chunk")[0]
                sections.add(section)

                nearestCand = None

                for candOcc in candOccs:
                    if nearestCand is None or abs(nearestCand[0] - nFirst) > abs(
                        candOcc[0] - nFirst
                    ):
                        nearestCand = candOcc

                for n in nearestCand[3]:
                    if n in highlights:
                        highlights[n] = "yellow"
                    else:
                        highlights[n] = "cyan"

                section = L.u(nearestCand[0], otype="chunk")[0]
                sections.add(section)

            sections = tuple((s,) for s in sorted(sections))

            s += len(sections)

            if asFile is None:
                app.table(sections, highlights=highlights, full=True)
            else:
                content[-1].append(
                    app.table(
                        sections, highlights=highlights, full=True, _asString=True
                    )
                )

                if i % 10 == 0:
                    app.info(f"{i:>4} variants done giving {s:>4} chunks")
                if i % 100 == 0:
                    content[-1].append(htmlEnd)
                    content.append([htmlStart])
                    empty = True

        app.info(f"{i:>4} matches done")

        if asFile is not None:
            content[-1].append(htmlEnd)
            if empty:
                content.pop()

            extraFileBase = f"{workDir}/extra"
            initTree(extraFileBase, fresh=True, gentle=True)

            for i, material in enumerate(content):
                extraFile = f"{extraFileBase}/{asFile}{i + 1:>02}.html"

                with fileOpen(extraFile, "w") as fh:
                    fh.write("\n".join(material))

                console(f"Extra triggers written to {extraFile}")

    def makeAlphabet(self):
        """Gathers the alphabet on which the corpus is based.

        The characters of the corpus have already been collected by Text-Fabric,
        and that is from where we pick them up.

        We separate the digits from the rest.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        app = self.app
        C = app.api.C
        workDir = self.workDir

        alphabetFile = f"{workDir}/alphabet.tsv"
        self.alphabetFile = alphabetFile

        with fileOpen(alphabetFile, "w") as fh:
            # This file will consist of one character per line,
            # for each distinct alpha character in the corpus, ordered by frequency.
            # Numeric characters will be put on a single line, with tabs in between.
            # All other characters will be ignored.

            digits = []

            for c, freq in C.characters.data["text-orig-full"]:
                if c.isalpha():
                    fh.write(f"{c}\n")
                elif c.isdigit():
                    digits.append(c)
            fh.write("\t".join(digits))

        console(f"Alphabet written to {alphabetFile}")

    def makeText(self):
        """Generate a plain text from the corpus.

        We make sure that we resolve the hyphenation of words across line
        boundaries.
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        NE = self.NE
        app = self.app
        api = app.api
        F = app.api.F
        L = app.api.L
        workDir = self.workDir

        rec = Recorder(api)

        lineType = self.settings.lineType
        slotType = F.otype.slotType
        maxSlot = F.otype.maxSlot
        lines = F.otype.s(lineType)
        lineEnds = {L.d(ln, otype=slotType)[-1] for ln in lines}
        skipTo = None

        for t in range(1, maxSlot + 1):
            tp = t + 1
            tpp = t + 2

            if tp in lineEnds and tp < maxSlot and F.str.v(tp) == "-":
                rec.start(t)
                rec.add(f"{F.str.v(t)}{F.after.v(t)}")
                rec.end(t)
                rec.start(tpp)
                rec.add(f"{F.str.v(tpp)}{F.after.v(tpp)}\n")
                rec.end(tpp)
                skipTo = tpp
            elif skipTo is not None:
                if t < skipTo:
                    continue
                else:
                    skipTo = None
            else:
                rec.start(t)
                rec.add(f"{F.str.v(t)}{F.after.v(t)}")
                rec.end(t)

        self.rec = rec
        textComplete = rec.text()
        self.textComplete = textComplete

        textFile = f"{workDir}/text.txt"

        with fileOpen(textFile, "w") as fh:
            fh.write(textComplete)

        NE.console(f"Text written to {textFile} - {len(textComplete)} characters")

    def makeLexicon(self):
        """Make a lexicon out of the triggers of a spreadsheet."""
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        NE = self.NE
        app = self.app
        workDir = self.workDir

        sheetData = self.sheetData
        NEinventory = sheetData.inventory

        app.indent(reset=True)
        app.info("Collecting the triggers for the lexicon")

        inventory = {}

        for eidkind, triggers in NEinventory.items():
            for trigger, scopes in triggers.items():
                inventory.setdefault(trigger, set())

                for occs in scopes.values():
                    for slots in occs:
                        inventory[trigger].add(tuple(slots))

        app.info(f"{len(inventory)} triggers collected")

        remSpaceRe = re.compile(r""" +([^A-Za-z])""")
        accentSpaceRe = re.compile(r"""([’']) +""")

        lexicon = {}
        mapNormal = {}

        lexiconOccs = {}
        self.lexiconOccs = lexiconOccs

        for name, occs in inventory.items():
            occStr = name
            occNormal = remSpaceRe.sub(r"\1", occStr)
            occNormal = accentSpaceRe.sub(r"\1", occNormal)
            nOccs = len(occs)
            lexicon[occNormal] = nOccs
            mapNormal[occNormal] = occStr
            lexiconOccs[occNormal] = occs

        sortedLexicon = sorted(lexicon.items(), key=lambda x: (-x[1], x[0].lower()))

        for name, n in sortedLexicon[0:10]:
            NE.console(f"  {n:>3} x {name}")

        NE.console("  ...")

        for name, n in sortedLexicon[-10:]:
            NE.console(f"  {n:>3} x {name}")

        NE.console(f"{len(lexicon):>8} lexicon length")

        lexiconFile = f"{workDir}/lexicon.tsv"
        self.lexiconFile = lexiconFile

        with fileOpen(lexiconFile, "w") as fh:
            for name, n in sorted(lexicon.items()):
                fh.write(f"{name}\t{n}\n")

        NE.console(f"Lexicon written to {lexiconFile}")

    def setupAnaliticcl(self):
        """Configure analiticcl for the big search.

        We gather the parameters from the variants section of the NER
        config file (see `tf.ner.settings`).

        For the description of the parameters, see the
        [analiticcl tutorial](https://github.com/proycon/analiticcl/blob/master/tutorial.ipynb)
        """
        if not self.properlySetup:
            console("This instance is not properly set up", error=True)
            return

        NE = self.NE
        VariantModel = self.VariantModel
        Weights = self.Weights
        ansettings = self.settings.analiticcl
        weights = ansettings.weights
        wld = weights.ld
        wlcs = weights.lcs
        wprefix = weights.prefix
        wsuffix = weights.suffix
        wcase = weights.case

        alphabetFile = self.alphabetFile
        lexiconFile = self.lexiconFile

        NE.console("Set up analiticcl")

        model = VariantModel(
            alphabetFile,
            Weights(ld=wld, lcs=wlcs, prefix=wprefix, suffix=wsuffix, case=wcase),
        )
        self.model = model
        model.read_lexicon(lexiconFile)
        model.build()



"""Annotation set management.

Annotation sets contain the annotations that the user generates by using
the tool.
"""

from ..core.generic import AttrDict
from ..core.helpers import console
from ..core.files import (
    fileExists,
    initTree,
    dirExists,
    dirContents,
    dirMake,
    dirCopy,
    dirRemove,
    dirMove,
)

from .data import Data
from .settings import ERROR, SET_ENT, SET_SHEET, SET_MAIN


class Sets(Data):
    def __init__(self, sets=None):
        """Methods to create, duplicate, rename and delete annotation sets.

        Annotation sets have names, given by the user.

        There is a special annotation set, whose name is the empty string,
        and whose content are the pre-existing entities, i.e. the entities that
        are present in the TF data as nodes and features.

        Users can not name sets with names that start with a dot.

        Annotation sets whose name start with a dot are generated by the system
        when a spreadsheets with entity triggers is processed.
        These sets are readonly, like the special annotation set, but they can
        be duplicated to ordinary sets. Those copies loose the relationship with
        the original spreadsheet.

        There is always one current annotation set, whose data is loaded into
        memory.

        Parameters
        ----------
        sets: object, optional None
            Entity sets to start with.
            If None, a fresh store of sets will be created by a parent class (Data).
        """
        Data.__init__(self, sets=sets)

        if not self.properlySetup:
            return

        browse = self.browse

        self.setName = ""
        """The current annotation set."""

        self.setInfo()

        self.setNames = set()
        """The set of names of annotation sets that are present on the file system."""

        self.readSets()

        if not browse:
            self.loadSetData()

    def setInfo(self, setName=None):
        """Give information about a set.

        Parameters
        ----------
        setName: string, optional None
            The name of the set we want info about. If `None`, give info about the
            current set.

        Returns
        -------
        tuple
            The info consists of: a representation of the name of the set; whether the
            set is readonly; whether the set corresponds to the baked-in entities of
            the dataset; whether the set corresponds to a spreadsheet.
        """
        settings = self.settings
        entitySet = settings.entitySet

        inObject = False

        if setName is None:
            setName = self.setName
            inObject = True

        setIsRo = setName == "" or setName.startswith(".")
        setIsSrc = setName == ""
        setIsX = setIsRo and not setIsSrc

        setNameRep = (
            f"{SET_ENT} {entitySet}"
            if setName == ""
            else (
                f"{SET_SHEET} " f"{setName[1:]}"
                if setName.startswith(".")
                else f"{SET_MAIN} {setName}"
            )
        )
        if inObject:
            self.setNameRep = setNameRep
            self.setIsRo = setIsRo
            self.setIsSrc = setIsSrc
            self.setIsX = setIsX
        else:
            return (setNameRep, setIsRo, setIsSrc, setIsX)

    def readSets(self):
        """Read the list current annotation sets (again).

        Use this when you change annotation sets outside the NER browser, e.g.
        by working with annotations in a Jupyter Notebook.
        """
        annoDir = self.annoDir

        self.setNames = set(dirContents(annoDir)[1])

    def getSetData(self):
        """Deliver the current set.

        Returns
        -------
        AttrDict
            A dictionary with the data of the set, with portions given by key.
            See `tf.ner.data.Data.fromSourceSet()` and `tf.ner.data.Data.processSet()`
        """
        setsData = self.sets
        setName = self.setName
        setData = setsData.setdefault(setName, AttrDict())
        return setData

    def setSet(self, newSetName):
        """Switch to a named annotation set.

        If the new set does not exist, it will be created.
        After the switch, the new set will be loaded into memory.

        Parameters
        ----------
        newSetName: string
            The name of the new annotation set to switch to.
        """
        if not self.properlySetup:
            return

        browse = self.browse

        if not browse:
            self.loadSetData()

        setNames = self.setNames
        setsData = self.sets
        setName = self.setName
        annoDir = self.annoDir
        newSetDir = f"{annoDir}/{newSetName}"

        (newSetNameRep, newSetRo, newSetSrc, newSetX) = self.setInfo(newSetName)

        if (not newSetSrc) and (newSetName not in setNames or not dirExists(newSetDir)):
            initTree(newSetDir, fresh=False)
            setNames.add(newSetName)

        if newSetName != setName:
            setName = newSetName
            self.setName = setName
            self.setInfo()
            self.loadSetData()

        if not browse:
            setNameRep = self.setNameRep
            entities = setsData[setName].entities
            nEntities = len(entities)
            plural = "" if nEntities == 1 else "s"
            self.console(
                f"Annotation set {setNameRep} " f"has {nEntities} annotation{plural}"
            )

    def _addToSet(self, newEntities):
        """Add a bunch of entities to the current set.

        Only for sets that correspond to sheets. This is to create such a set,
        it is not meant to call this function manually in a Jupyter notebook.

        Parameters
        ----------
        newSetName: string
            The name of the new annotation set to switch to.
        """
        if not self.properlySetup:
            return

        setIsX = self.setIsX

        if not setIsX:
            return

        self._clearSetData()
        self.addEntities(newEntities, returns=False, _lowlevel=True)

    def resetSet(self):
        """Clear the current annotation set.

        The special set `""` cannot be reset, because it is read-only.
        """
        if not self.properlySetup:
            return

        settings = self.settings
        setName = self.setName
        setIsRo = self.setIsRo
        entitySet = settings.entitySet

        if setIsRo:
            console(f"Resetting the {entitySet} has no effect")
            return

        browse = self.browse

        setsData = self.sets
        annoDir = self.annoDir
        setDir = f"{annoDir}/{setName}"

        initTree(setDir, fresh=True, gentle=True)
        self.loadSetData()

        if not browse:
            setNameRep = self.setNameRep
            entities = setsData[setName].entities
            nEntities = len(entities)
            plural = "" if nEntities == 1 else "s"
            self.console(
                f"Annotation set {setNameRep} has {nEntities} annotation{plural}"
            )

    def setDup(self, dupSet):
        """Duplicates the current set to a set with a new name.

        !!! hint "The readonly sets can be duplicated"
            After duplication of a read-only set, the duplicate
            copy is modifiable.
            In this way you can make corrections to the set of pre-existing,
            tool-generated annotations.

        The current set changes to the result of the duplication.

        Parameters
        ----------
        dupSet: string
            The name of new set that is the result of the duplication.

        Returns
        -------
        list
            A list of messages. Every message is a tuple of kind and content.
        """
        if not self.properlySetup:
            return []

        setNames = self.setNames
        setsData = self.sets
        setName = self.setName
        setIsSrc = self.setIsSrc
        annoDir = self.annoDir
        annoPath = f"{annoDir}/{dupSet}"

        messages = []

        if dupSet in setNames:
            messages.append((ERROR, f"""Set {dupSet} already exists"""))
        else:
            if setIsSrc:
                dataFile = f"{annoPath}/entities.tsv"

                if fileExists(dataFile):
                    messages.append((ERROR, f"""Set {dupSet} already exists"""))
                else:
                    dirMake(annoPath)
                    self.saveEntitiesAs(dataFile)
                    setNames.add(dupSet)
                    setsData[dupSet] = setsData[setName]
                    self.setName = dupSet
                    self.setInfo()
            else:
                if not dirCopy(
                    f"{annoDir}/{setName}",
                    annoPath,
                    noclobber=True,
                ):
                    messages.append(
                        (ERROR, f"""Could not copy {setName} to {dupSet}""")
                    )
                else:
                    setNames.add(dupSet)
                    setsData[dupSet] = setsData[setName]
                    self.setName = dupSet
                    self.setInfo()

        return messages

    def setDel(self, delSet):
        """Remove a named set.

        If the removed set happens to be the current set, the current set changes
        to the special set named `""`.

        Parameters
        ----------
        delSet: string
            The name of the set to be removed.
            It is not allowed to remove the special set named `""`.

        Returns
        -------
        list
            A list of messages. Every message is a tuple of kind and content.
        """
        if not self.properlySetup:
            return []

        messages = []
        (delSetRep, delSetRo, delSetSrc, delSetX) = self.setInfo(setName=delSet)

        if delSetRo:
            messages.append(
                (ERROR, f"""Cannot remove set {delSetRep} because it is read-only""")
            )
            return messages

        setNames = self.setNames
        setsData = self.sets
        annoDir = self.annoDir
        annoPath = f"{annoDir}/{delSet}"

        dirRemove(annoPath)

        if dirExists(annoPath):
            messages.append((ERROR, f"""Could not remove {delSetRep}"""))
        else:
            setNames.discard(delSet)
            del setsData[delSet]
            if self.setName == delSet:
                self.setName = ""
                self.setInfo()

        return messages

    def setMove(self, moveSet):
        """Renames a named set.

        The current set changes to the renamed set.
        It is not possible to rename the special set named `""`.
        It is also forbidden to rename another set to the special set.

        Parameters
        ----------
        moveSet: string
            The new name of the current set.

        Returns
        -------
        list
            A list of messages. Every message is a tuple of kind and content.
        """
        if not self.properlySetup:
            return []

        messages = []
        (moveSetRep, moveSetRo, moveSetSrc, moveSetX) = self.setInfo(setName=moveSet)

        if moveSetRo:
            messages.append((ERROR, f"""Cannot rename a set to ""{moveSetRep}"""))
            return messages

        setName = self.setName
        setNameRep = self.setNameRep
        setIsRo = self.setIsRo

        if setIsRo:
            messages.append((ERROR, f"""Cannot rename set ""{setNameRep}"""))
            return messages

        setNames = self.setNames
        setsData = self.sets
        annoDir = self.annoDir
        annoPath = f"{annoDir}/{moveSet}"

        if dirExists(annoPath):
            messages.append((ERROR, f"""Set {moveSetRep} already exists"""))
        else:
            if not dirMove(f"{annoDir}/{setName}", annoPath):
                messages.append(
                    (
                        ERROR,
                        f"""Could not rename {setNameRep} to {moveSetRep}""",
                    )
                )
            else:
                setNames.add(moveSet)
                setNames.discard(setName)
                setsData[moveSet] = setsData[setName]
                del setsData[setName]
                self.setName = moveSet
                self.setInfo()

        return messages



"""
Command-line argument processing
"""

from zlib import crc32

from ..parameters import PORT_BASE
from ..core.files import getLocation

# COMMAND-LINE ARGS


def getPort(slug):
    portOffset = crc32(slug.encode("utf8")) % 10000
    return PORT_BASE + portOffset


def argApp(cargs, simple):
    (appName, checkoutApp, dataLoc) = argParam(cargs)
    backend = _argCollect("backend", cargs)
    checkout = _argCollect("checkout", cargs)
    version = _argCollect("version", cargs)
    dataRelative = _argCollect("relative", cargs)

    if (
        appName is None
        and checkoutApp is None
        and dataLoc is None
        and backend is None
        and checkout is None
    ):
        (backend, org, repo, relative) = getLocation()
        appName = f"{org}/{repo}{relative}"
        checkoutApp = "clone"
        checkout = "clone"

    if simple:
        return f"{appName}:{version}"

    locations = _argCollect("locations", cargs)
    modules = _argCollect("modules", cargs)
    moduleRefs = _argCollect("mod", cargs)
    version = _argCollect("version", cargs)
    setFile = _argCollect("sets", cargs)

    return (
        None
        if appName is None
        else dict(
            appName=appName,
            backend=backend,
            checkoutApp=checkoutApp,
            dataLoc=dataLoc,
            checkout=checkout,
            relative=dataRelative,
            locations=locations,
            modules=modules,
            moduleRefs=moduleRefs,
            setFile=setFile,
            version=version,
        )
    )


def argNoweb(cargs):
    for arg in cargs:
        if arg == "-noweb":
            return True
    return False


def _argCollect(prefix, cargs):
    for arg in cargs:
        if arg.startswith(f"--{prefix}="):
            return arg[len(prefix) + 3 :]
    return None


def argParam(cargs):
    appName = None
    checkoutApp = None

    for arg in cargs:
        if arg.startswith("-"):
            continue
        appName = arg
        break

    if appName is None:
        return (None, None, None)

    if appName.startswith("data:"):
        dataLoc = appName[5:]
        appName = None
        checkoutApp = None
    elif appName.startswith("app:"):
        dataLoc = None
        checkoutApp = None
    else:
        dataLoc = None
        parts = appName.split(":", maxsplit=1)
        if len(parts) == 1:
            parts.append("")
        (appName, checkoutApp) = parts

    return (appName, checkoutApp, dataLoc)



import sys
import os

from subprocess import call, Popen, PIPE

import errno
import time
import unicodedata

from tf.core.files import fileOpen

REMOTE = "origin"
BRANCH = "gh-pages"
SITE = "site"

# COPIED FROM MKDOCS AND MODIFIED


def console(*args):
    sys.stdout.write(" ".join(args) + "\n")
    sys.stdout.flush()


def _enc(text):
    if isinstance(text, bytes):
        return text
    return text.encode()


def _dec(text):
    if isinstance(text, bytes):
        return text.decode("utf-8")
    return text


def _write(pipe, data):
    try:
        pipe.stdin.write(data)
    except OSError as e:
        if e.errno != errno.EPIPE:
            raise


def _normalize_path(path):
    # Fix UNICODE pathnames on macos
    # See: https://stackoverflow.com/a/5582439/44289
    if sys.platform == "darwin":
        return unicodedata.normalize("NFKC", _dec(path))
    return path


def _try_rebase(remote, branch):
    cmd = ["git", "rev-list", "--max-count=1", "{}/{}".format(remote, branch)]
    p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    (rev, _) = p.communicate()
    if p.wait() != 0:
        return True
    cmd = ["git", "update-ref", "refs/heads/%s" % branch, _dec(rev.strip())]
    if call(cmd) != 0:
        return False
    return True


def _get_config(key):
    p = Popen(["git", "config", key], stdin=PIPE, stdout=PIPE)
    (value, _) = p.communicate()
    return value.decode("utf-8").strip()


def _get_prev_commit(branch):
    cmd = ["git", "rev-list", "--max-count=1", branch, "--"]
    p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    (rev, _) = p.communicate()
    if p.wait() != 0:
        return None
    return rev.decode("utf-8").strip()


def _mk_when(timestamp=None):
    if timestamp is None:
        timestamp = int(time.time())
    currtz = "%+05d" % (-1 * time.timezone / 36)  # / 3600 * 100
    return "{} {}".format(timestamp, currtz)


def _start_commit(pipe, branch, message):
    uname = _dec(_get_config("user.name"))
    email = _dec(_get_config("user.email"))
    _write(pipe, _enc("commit refs/heads/%s\n" % branch))
    _write(pipe, _enc("committer {} <{}> {}\n".format(uname, email, _mk_when())))
    _write(pipe, _enc("data %d\n%s\n" % (len(message), message)))
    head = _get_prev_commit(branch)
    if head:
        _write(pipe, _enc("from %s\n" % head))
    _write(pipe, _enc("deleteall\n"))


def _add_file(pipe, srcpath, tgtpath):
    with fileOpen(srcpath, mode="rb") as handle:
        if os.access(srcpath, os.X_OK):
            _write(pipe, _enc("M 100755 inline %s\n" % tgtpath))
        else:
            _write(pipe, _enc("M 100644 inline %s\n" % tgtpath))
        data = handle.read()
        _write(pipe, _enc("data %d\n" % len(data)))
        _write(pipe, _enc(data))
        _write(pipe, _enc("\n"))


def _add_nojekyll(pipe):
    _write(pipe, _enc("M 100644 inline .nojekyll\n"))
    _write(pipe, _enc("data 0\n"))
    _write(pipe, _enc("\n"))


def _gitpath(fname):
    norm = os.path.normpath(fname)
    return "/".join(norm.split(os.path.sep))


def _ghp_import():
    if not _try_rebase(REMOTE, BRANCH):
        print("Failed to rebase %s branch.", BRANCH)

    console(f"copy docs to the {BRANCH} branch")
    cmd = ["git", "fast-import", "--date-format=raw", "--quiet"]
    kwargs = {"stdin": PIPE}
    if sys.version_info >= (3, 2, 0):
        kwargs["universal_newlines"] = False
    pipe = Popen(cmd, **kwargs)
    _start_commit(pipe, BRANCH, "docs update")
    for path, _, fnames in os.walk(SITE):
        for fn in fnames:
            fpath = os.path.join(path, fn)
            fpath = _normalize_path(fpath)
            gpath = _gitpath(os.path.relpath(fpath, start=SITE))
            _add_file(pipe, fpath, gpath)
    _add_nojekyll(pipe)
    _write(pipe, _enc("\n"))
    pipe.stdin.close()
    if pipe.wait() != 0:
        sys.stdout.write(_enc("Failed to process commit.\n"))

    console(f"push {BRANCH} branch to GitHub")
    cmd = ["git", "push", REMOTE, BRANCH]
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE)
    (out, err) = proc.communicate()
    result = proc.wait() == 0

    return result, _dec(err)


def deploy(org, repo):
    (result, error) = _ghp_import()
    if not result:
        print("Failed to deploy to GitHub with error: \n%s", error)
        raise SystemExit(1)
    else:
        url = f"https://{org}.github.io/{repo}"
        print("Your documentation should shortly be available at: " + url)



"""
# Transcription

TF has support for several writing systems, by means of
transcription tables and fonts that will be invoked when displaying the main text.

It also calls functions to use these tables for converting Hebrew and Syriac
text material to transliterated representations and back.

There is also a phonetic transcription for Hebrew, designed in
[phono.ipynb](https://nbviewer.jupyter.org/github/etcbc/phono/blob/master/programs/phono.ipynb)

## Character tables and fonts

### `hbo` Hebrew

`tf.writing.hebrew`: full list of characters covered by the ETCBC and phonetic transcriptions

Font `Ezra SIL`.

### `syc` Syriac

`tf.writing.syriac`: full list of characters covered by the ETCBC transcriptions

Font `Estrangelo Edessa`.

### `ara` Arabic

`tf.writing.arabic`: full list of characters covered by the transcription used for the Quran

Font `AmiriQuran`.

### `grc` Greek

Font `Gentium`.

### `akk` Akkadian

Font `Santakku`.

### `uga` Ugaritic

Font `Santakku`.

### `cld` Neo Aramaic

Font `CharisSIL-R`.
"""

import re


class Transcription:
    """Conversion between UNICODE and various transcriptions.

    Usage notes:

    Invoke the transcription functionality as follows:

    ```
    from tf.writing.transcription import Transcription
    ```

    Some of the attributes and methods below are *class* attributes,
    others are instance attributes.

    A class attribute `aaa` can be retrieved by saying

    ``` python
    Transcription.aaa
    ```

    To retrieve an instance attribute, you need an instance first, like

    ``` python
    tr = Transcription()
    tr.aaa
    ```
    """

    decomp = {
        "\u05E9\u05C1": "\uFB2A",
        "\u05E9\u05C2": "\uFB2B",
    }
    hebrew_mapping = {
        "_": " ",  # space inside word
        "92": "\u0591",  # etnahta = atnach
        "01": "\u0592",  # segolta
        "65": "\u0593",  # shalshelet
        "80": "\u0594",  # zaqef_qatan
        "85": "\u0595",  # zaqef_gadol
        "73": "\u0596",  # tipeha = tifcha
        "81": "\u0597",  # revia = rebia
        "82": "\u0598",  # zarqa = tsinorit = zinorit = sinnorit
        "03": "\u0599",  # pashta
        "10": "\u059A",  # yetiv = yetib
        "91": "\u059B",  # tevir = tebir
        "61": "\u059C",  # geresh
        "11": "\u059D",  # geresh muqdam = mugrash
        "62": "\u059E",  # gershayim = garshayim
        "84": "\u059F",  # qarney para = pazer_gadol
        "14": "\u05A0",  # telisha_gedola
        "44": "\u05A0",  # telisha_gedola = telisha_gedola_med
        "83": "\u05A1",  # pazer
        "74": "\u05A3",  # munah = munach
        "70": "\u05A4",  # mahapakh = mehuppach
        "71": "\u05A5",  # merkha = merecha
        "72": "\u05A6",  # merkha kefula = merecha_kepula
        "94": "\u05A7",  # darga
        "63": "\u05A8",  # qadma = azla
        "33": "\u05A8",  # pashta_med < qadma
        "04": "\u05A9",  # telisha_qetana
        "24": "\u05A9",  # telisha_qetana = telisha_qetana_med
        "93": "\u05AA",  # yera ben yomo = yerach
        "60": "\u05AB",  # ole = ole_weyored
        "64": "\u05AC",  # iluy = illuy
        "13": "\u05AD",  # dehi = dechi
        "02": "\u05AE",  # zinor = sinnor
        "*": "\u05AF",  # masora circle
        ":": "\u05B0",  # sheva = shewa
        ":E": "\u05B1",  # hataf segol = chataph_segol
        ":A": "\u05B2",  # hataf patah = chataph_patach
        ":@": "\u05B3",  # hataf qamats = chataph_qamats
        "I": "\u05B4",  # hiriq = chiriq
        ";": "\u05B5",  # tsere
        "E": "\u05B6",  # segol
        "A": "\u05B7",  # patach
        "@": "\u05B8",  # qamats
        "O": "\u05B9",  # holam = cholam
        "U": "\u05BB",  # qubuts = qubbuts
        ".": "\u05BC",  # dagesh
        "25": "\u05BD",  # silluq yamin
        "45": "\u05BD",  # meteg
        "35": "\u05BD",  # meteg (tikon)
        "75": "\u05BD",  # siluq = silluq
        "95": "\u05BD",  # meteg = meteg_yamin
        "&": "\u05BE",  # maqaf
        ",": "\u05BF",  # rafe = raphe
        "05": "\u05C0",  # paseq
        ".c": "\u05C1",  # shin dot
        ".f": "\u05C2",  # sin dot
        "00": "\u05C3",  # sof_pasuq
        "52": "\u05C4",  # upper dot = puncta_above
        "53": "\u05C5",  # lower dot = puncta_below
        "ñ": "\u05C6\u0307",  # nun hafukha
        "Ñ": "\u05C6\u0307",  # nun hafukha
        ">": "\u05D0",  # alef
        "B": "\u05D1",  # bet
        "G": "\u05D2",  # gimel
        "D": "\u05D3",  # dalet
        "H": "\u05D4",  # he
        "W": "\u05D5",  # vav
        "Z": "\u05D6",  # zayin
        "X": "\u05D7",  # het
        "V": "\u05D8",  # tet
        "J": "\u05D9",  # yod
        "k": "\u05DA",  # kaf final
        "K": "\u05DB",  # kaf
        "L": "\u05DC",  # lamed
        "m": "\u05DD",  # mem final
        "M": "\u05DE",  # mem
        "n": "\u05DF",  # nun final
        "N": "\u05E0",  # nun
        "S": "\u05E1",  # samekh
        "<": "\u05E2",  # ayin
        "p": "\u05E3",  # pe final
        "P": "\u05E4",  # pe
        "y": "\u05E5",  # tsadi final
        "Y": "\u05E6",  # tsadi
        "Q": "\u05E7",  # qof
        "R": "\u05E8",  # resh
        "#": "\u05E9",  # sin unpointed
        "T": "\u05EA",  # tav
        "C": "\uFB2A",  # shin pointed
        "F": "\uFB2B",  # sin pointed
        "55": "<UNMAPPED 55=large letter>",  # large_letter
        "56": "<UNMAPPED 56=small letter>",  # small_letter
        "57": "<UNMAPPED 57=suspended letter>",  # suspended_letter
        "-": "",  # suppress space afterward
        "'": "\u05f3",  # punctuation geresh
        '"': "\u05f4",  # punctuation gershayim
    }
    """
    Maps all ETCBC transliteration character combinations for Hebrew to UNICODE.

    Example: sof-pasuq:

    ``` python
    Transcription.hebrew_mapping['00']
    ```

    Output:

    ```
    ׃
    ```
    """

    hebrew_cons = ">BGDHWZXVJKLMNS<PYQRFCT"
    trans_final_pat = re.compile(
        r"(["
        + hebrew_cons
        + r"][^_&]*)([KMNPY])([^"
        + hebrew_cons
        + r"_&]*(:?[_&]|\Z))"
    )
    trans_hebrew_pat = re.compile(r"(:[AE@]|.[cf]|:|[0-9][0-9]|.)")
    swap_accent_pat = re.compile(
        r"(\A|[_&])([0-9][0-9])([" + hebrew_cons + r"])([:;,.EAIOU@]*)"
    )
    remove_accent_pat = re.compile(r"((?:[0-9][0-9])|[,*])")
    remove_point_pat = re.compile(
        r"((?:[0-9][0-9])|(?:\.[cf])|(?::[@AE])|[,.:;@AEIOU*])"
    )
    remove_psn_pat = re.compile(r"00[ _SPNÑñ]*")
    remove_psq_pat = re.compile(r"(?:[ _]+05[ _]*)|(?:05[ _]+)")
    shin_pat = re.compile(r"[CF]")
    ph_simple_pat = re.compile(r"([ˈˌᵊᵃᵒᵉāo*])")
    noorigspace = re.compile(
        r"""
          (?: [&-]\Z)           # space, maqef or nospace
        | (?:
               0[05]            # sof pasuq or paseq
               (?:_[SNP])*      # nun hafukha, setumah, petuhah at end of verse
               \Z
          )
        | (?:_[SPN])+           #  nun hafukha, setumah, petuhah between words
    """,
        re.X,
    )

    ugaritic_mappingi = {
        "𐎀": "a",
        "𐎁": "b",
        "𐎂": "g",
        "𐎃": "ḫ",
        "𐎄": "d",
        "𐎅": "h",
        "𐎆": "w",
        "𐎇": "z",
        "𐎈": "ḥ",
        "𐎉": "ṭ",
        "𐎊": "y",
        "𐎋": "k",
        "𐎌": "š",
        "𐎍": "l",
        "𐎎": "m",
        "𐎏": "ḏ",
        "𐎐": "n",
        "𐎑": "ẓ",
        "𐎒": "s",
        "𐎓": "ˤ",
        "𐎔": "p",
        "𐎕": "ṣ",
        "𐎖": "q",
        "𐎗": "r",
        "𐎘": "ṯ",
        "𐎙": "ġ",
        "𐎚": "t",
        "𐎛": "i",
        "𐎜": "u",
        "𐎝": "s2",
        "𐎟": ".",
    }
    """
    Maps Ugaritic unicode characters to their conventional transliteration characters.

    Unidentified characters:

    * `x` (damaged ?)
    * `/` (alternative ?) only twice, in `atypˤtba/r` and `xxxxl/d…`
    * `,` (comma) only once in a very long word starting at 551 ... `km,ad` ...
    * `<` `>` (brackets marking uncertainty ?)
    * `…` (unreadable ?)
    * `00a0` (non-breaking space)
    """

    ugaritic_mapping = {v: k for (k, v) in ugaritic_mappingi.items()}

    syriac_mapping_simple = {
        ">": "\u0710",  # alaph
        "B": "\u0712",  # beth
        "G": "\u0713",  # gamal
        "D": "\u0715",  # dalat
        "H": "\u0717",  # he
        "W": "\u0718",  # waw
        "Z": "\u0719",  # zain
        "X": "\u071A",  # heth
        "V": "\u071B",  # teth
        "J": "\u071D",  # yudh
        "K": "\u071F",  # kaph
        "L": "\u0720",  # lamadh
        "M": "\u0721",  # mim
        "N": "\u0722",  # nun
        "S": "\u0723",  # semkath
        "<": "\u0725",  # e
        "P": "\u0726",  # pe
        "Y": "\u0728",  # sadhe
        "Q": "\u0729",  # qaph
        "R": "\u072A",  # rish
        "C": "\u072B",  # shin
        "T": "\u072C",  # taw
        "s": "\u0724",  # semkath final
        "p": "\u0727",  # pe reversed
    }

    syriac_mapping_pil = {
        # LETTERS
        "'": "\u0710",  # alaph
        "b": "\u0712",  # beth
        "g": "\u0713",  # gamal
        "d": "\u0715",  # dalat
        "h": "\u0717",  # he
        "w": "\u0718",  # waw
        "z": "\u0719",  # zain
        "H": "\u071A",  # heth
        "T": "\u071B",  # teth
        "y": "\u071D",  # yod
        "k": "\u071F",  # kaf
        "l": "\u0720",  # lamad
        "m": "\u0721",  # mim
        "n": "\u0722",  # nun
        "s": "\u0723",  # semkath
        "`": "\u0725",  # 'e
        "p": "\u0726",  # pe
        "S": "\u0728",  # tsade
        "q": "\u0729",  # qof
        "r": "\u072A",  # resh
        "$": "\u072B",  # shin
        "t": "\u072C",  # taw
        # WORD-BOUND DIACRITICS
        '"': "\u0308",  # seyame
        "#": "\u0323",  # diacritical dot below
        "^": "\u0307",  # diacritical dot above
        "~": "\u0307",  # abbreviation mark
        # NON-VOCALIC LETTER-BOUND DIACRITICS
        "#,": "\u0742",  # rukkakha
        '#"': "\u0342",  # unclear (COMBINING DIAERESIS BELOW)
        "#!": "\u0744",  # unclear (SYRIAC TWO VERTICAL DOTS BELOW)
        "#_": "\u0331",  # linea occultans infera
        "^,": "\u0741",  # qushshaya
        "^!": "\u0743",  # unclear (SYRIAC TWO VERTICAL DOTS ABOVE)
        "^_": "\u0304",  # linea occultans supera
        # VOCALIC LETTER-BOUND DIACRITICS
        ":": "",  # shewa
        "A": "\u0733",  # qamets
        "A1": "\u0734",  # zeqapa
        "A2": "\u0735",  # zeqofo
        "E": "\u0739",  # tsere, revasa karya
        "O": "\u073F",  # holem, rewaha
        "a": "\u0730",  # patah
        "a1": "\u0731",  # petaha
        "a2": "\u0732",  # petoho
        "e": "\u0736",  # segol
        "e1": "\u0737",  # revasa arrika
        "e2": "\u0738",  # revoso
        "i": "\u073A",  # hireq
        "i1": "\u073B",  # hevoso
        "y#": "\u071D\u073C",  # hevasa
        "u": "\u073D",  # qubbuts
        "u1": "\u073E",  # esoso
        "w#": "\u0718\u073C",  # esasa allisa
        "w^": "\u0718\u073F",  # esasa rewiha
        # INTERPUNCTION
        "#.": "\u0702",  # menachta, meshalyana (ES), metdamrana, samka
        "#:": "\u0704",  # metkashpana (ES)
        "#\\": "\u0709",  # tahtaya, metkashpana (WS), meshalyana (WS)
        "=.": "\u002E",  # pasuqa
        "=/": "\u0707",  # elaya
        "=:": "\u003A",  # shewaya (WS), zauga (ES)
        "=\\": "\u0706",  # unclear (SYRIAC COLON SKEWED LEFT)
        "^.": "\u0701",  # paquda, metkashpana (ES), meshalyana (ES), etsyana, meshalana?
        '^"': "\u0705",  # rahta
        "^:": "\u0703",  # taksa (WS), zauga elaya (ES)
        "^\\": "\u0708",  # unclear (SYRIAC SUPRALINEAR COLON SKEWED LEFT)
        # PERICOPE MARKERS
        "*": "\u0700",  # rosette
        ".": "\u00B7",  # common dot in caesuras
        "@": "\u2722",  # vignette
        "_": "\u2014",  # dash in caesuras
        "o": "\u2022",  # large dot in caesuras
    }

    syriac_mapping = {  # this is WIT
        # LETTERS
        ">": "\u0710",  # alaph
        "B": "\u0712",  # beth
        "G": "\u0713",  # gamal
        "D": "\u0715",  # dalat
        "H": "\u0717",  # he
        "W": "\u0718",  # waw
        "Z": "\u0719",  # zain
        "X": "\u071A",  # heth
        "V": "\u071B",  # teth
        "J": "\u071D",  # yod
        "K": "\u071F",  # kaf
        "L": "\u0720",  # lamad
        "M": "\u0721",  # mim
        "N": "\u0722",  # nun
        "S": "\u0723",  # semkath
        "<": "\u0725",  # 'e
        "P": "\u0726",  # pe
        "Y": "\u0728",  # tsade
        "Q": "\u0729",  # qof
        "R": "\u072A",  # resh
        "C": "\u072B",  # shin
        "T": "\u072C",  # taw
        # WORD-BOUND DIACRITICS
        '"': "\u0308",  # seyame
        "#": "\u0323",  # diacritical dot below
        "^": "\u0307",  # diacritical dot above
        # NON-VOCALIC LETTER-BOUND DIACRITICS
        "^!": "\u0743",  # unclear (SYRIAC TWO VERTICAL DOTS ABOVE)
        # VOCALIC LETTER-BOUND DIACRITICS
        ":": "",  # shewa
        "A": "\u0733",  # qamets
        "A1": "\u0734",  # zeqapa
        "A2": "\u0735",  # zeqofo
        "O": "\u073F",  # holem, rewaha
        "@": "\u0730",  # patah
        "@1": "\u0731",  # petaha
        "@2": "\u0732",  # petoho
        "E": "\u0736",  # segol
        "E1": "\u0737",  # revasa arrika
        "E2": "\u0738",  # revoso
        "I": "\u073A",  # hireq
        "I1": "\u073B",  # hevoso
        "U": "\u073D",  # qubbuts
        "U1": "\u073E",  # esoso
        # INTERPUNCTION
        "#\\": "\u0709",  # tahtaya, metkashpana (WS), meshalyana (WS)
        "=.": "\u002E",  # pasuqa
        "=#": "\u0707",  # elaya
        "=:": "\u003A",  # shewaya (WS), zauga (ES)
        "=^": "\u0706",  # unclear (SYRIAC COLON SKEWED LEFT)
        "=/": "\u0707",  # elaya
        "=\\": "\u0706",  # unclear (SYRIAC COLON SKEWED LEFT)
        "^:": "\u0703",  # taksa (WS), zauga elaya (ES)
        "^\\": "\u0708",  # unclear (SYRIAC SUPRALINEAR COLON SKEWED LEFT)
        # PERICOPE MARKERS
        "*": "\u0700",  # rosette
        ".": "\u00B7",  # common dot in caesuras
        "_": "\u2014",  # dash in caesuras
        "o": "\u2022",  # large dot in caesuras
    }
    """
    Maps all ETCBC transliteration character combinations for Syriac to UNICODE.

    Example: semkath-final:

    ``` python
    Transcription.syriac_mapping['s']
    ```

    Output:

    ```
    ܤ
    ```
    """

    trans_syriac_pat = re.compile(r"([AE@IU][12]?|=[.#:\^/\\]|[\^#][!:\\]|.)")

    arabic_mapping = {
        " ": "\u0020",  # SPACE
        "'": "\u0621",  # ARABIC LETTER HAMZA
        ">": "\u0623",  # ARABIC LETTER ALEF WITH HAMZA ABOVE
        "&": "\u0624",  # ARABIC LETTER WAW WITH HAMZA ABOVE
        "<": "\u0625",  # ARABIC LETTER ALEF WITH HAMZA BELOW
        "}": "\u0626",  # ARABIC LETTER YEH WITH HAMZA ABOVE
        "A": "\u0627",  # ARABIC LETTER ALEF
        "b": "\u0628",  # ARABIC LETTER BEH
        "p": "\u0629",  # ARABIC LETTER TEH MARBUTA
        "t": "\u062a",  # ARABIC LETTER TEH
        "v": "\u062b",  # ARABIC LETTER THEH
        "j": "\u062c",  # ARABIC LETTER JEEM
        "H": "\u062d",  # ARABIC LETTER HAH
        "x": "\u062e",  # ARABIC LETTER KHAH
        "d": "\u062f",  # ARABIC LETTER DAL
        "*": "\u0630",  # ARABIC LETTER THAL
        "r": "\u0631",  # ARABIC LETTER REH
        "z": "\u0632",  # ARABIC LETTER ZAIN
        "s": "\u0633",  # ARABIC LETTER SEEN
        "$": "\u0634",  # ARABIC LETTER SHEEN
        "S": "\u0635",  # ARABIC LETTER SAD
        "D": "\u0636",  # ARABIC LETTER DAD
        "T": "\u0637",  # ARABIC LETTER TAH
        "Z": "\u0638",  # ARABIC LETTER ZAH
        "E": "\u0639",  # ARABIC LETTER AIN
        "g": "\u063a",  # ARABIC LETTER GHAIN
        "_": "\u0640",  # ARABIC TATWEEL
        "f": "\u0641",  # ARABIC LETTER FEH
        "q": "\u0642",  # ARABIC LETTER QAF
        "k": "\u0643",  # ARABIC LETTER KAF
        "l": "\u0644",  # ARABIC LETTER LAM
        "m": "\u0645",  # ARABIC LETTER MEEM
        "n": "\u0646",  # ARABIC LETTER NOON
        "h": "\u0647",  # ARABIC LETTER HEH
        "w": "\u0648",  # ARABIC LETTER WAW
        "Y": "\u0649",  # ARABIC LETTER ALEF MAKSURA
        "y": "\u064a",  # ARABIC LETTER YEH
        "F": "\u064b",  # ARABIC FATHATAN
        "N": "\u064c",  # ARABIC DAMMATAN
        "K": "\u064d",  # ARABIC KASRATAN
        "a": "\u064e",  # ARABIC FATHA
        "u": "\u064f",  # ARABIC DAMMA
        "i": "\u0650",  # ARABIC KASRA
        "~": "\u0651",  # ARABIC SHADDA
        "o": "\u0652",  # ARABIC SUKUN
        "^": "\u0653",  # ARABIC MADDAH ABOVE
        "#": "\u0654",  # ARABIC HAMZA ABOVE
        "`": "\u0670",  # ARABIC LETTER SUPERSCRIPT ALEF
        "{": "\u0671",  # ARABIC LETTER ALEF WASLA
        ":": "\u06dc",  # ARABIC SMALL HIGH SEEN
        "@": "\u06df",  # ARABIC SMALL HIGH ROUNDED ZERO
        '"': "\u06e0",  # ARABIC SMALL HIGH UPRIGHT RECTANGULAR ZERO
        "[": "\u06e2",  # ARABIC SMALL HIGH MEEM ISOLATED FORM
        ";": "\u06e3",  # ARABIC SMALL LOW SEEN
        ",": "\u06e5",  # ARABIC SMALL WAW
        ".": "\u06e6",  # ARABIC SMALL YEH
        "!": "\u06e8",  # ARABIC SMALL HIGH NOON
        "-": "\u06ea",  # ARABIC EMPTY CENTRE LOW STOP
        "+": "\u06eb",  # ARABIC EMPTY CENTRE HIGH STOP
        "%": "\u06ec",  # ARABIC ROUNDED HIGH STOP WITH FILLED CENTRE
        "]": "\u06ed",  # ARABIC SMALL LOW MEEM
    }
    """
    Maps an Arabic transliteration character to UNICODE.
    This is the mapping used in the Quran representation on tanzil.net.

    Example: beh

    ``` python
    Transcription.syriac_mapping['b']
    ```

    Output:

    ```
    ب
    ```

    Maps an Arabic letter in UNICODE to its transliteration

    Example: beh transliteration

    ``` python
    Transcription.syriac_mapping['ب']
    ```

    Output:

    ```
    b
    ```
    """

    arabic_mappingi = {v: k for (k, v) in arabic_mapping.items()}

    arabicTrans = {
        "\u0020": (" ", " ", " ", " "),  # SPACE
        "\u060c": ("‚", ",", ",", ","),  # ARABIC COMMA
        "\u061b": ("„", ";", ";", ";"),  # ARABIC SEMICOLON
        "\u061f": ("?", "?", "?", "?"),  # ARABIC QUESTION MARK
        "\u0621": ("'", "'", "ʾ", "'"),  # ARABIC LETTER HAMZA
        "\u0622": ("»", "_a", "ā", "ā"),  # ARABIC LETTER ALEF WITH MADDA ABOVE
        "\u0623": (">", "_a", "ā", "ā"),  # ARABIC LETTER ALEF WITH HAMZA ABOVE
        "\u0624": ("&", "'", "ʾ", "'"),  # ARABIC LETTER WAW WITH HAMZA ABOVE
        "\u0625": ("<", "_a", "ā", "ā"),  # ARABIC LETTER ALEF WITH HAMZA BELOW
        "\u0626": ("}", "'", "ʾ", "y"),  # ARABIC LETTER YEH WITH HAMZA ABOVE
        "\u0627": ("A", "_a", "ā", "ā"),  # ARABIC LETTER ALEF
        "\u0628": ("b", "b", "b", "b"),  # ARABIC LETTER BEH
        "\u0629": ("p", "=t", "ŧ", "t"),  # ARABIC LETTER TEH MARBUTA
        "\u062a": ("t", "t", "t", "t"),  # ARABIC LETTER TEH
        "\u062b": ("v", "_t", "ṯ", "th"),  # ARABIC LETTER THEH
        "\u062c": ("j", "j", "ǧ", "j"),  # ARABIC LETTER JEEM
        "\u062d": ("H", "*h", "ḥ", "ḥ"),  # ARABIC LETTER HAH
        "\u062e": ("x", "_h", "ḫ", "kh"),  # ARABIC LETTER KHAH
        "\u062f": ("d", "d", "d", "d"),  # ARABIC LETTER DAL
        "\u0630": ("*", "_d", "ḏ", "dh"),  # ARABIC LETTER THAL
        "\u0631": ("r", "r", "r", "r"),  # ARABIC LETTER REH
        "\u0632": ("z", "z", "z", "z"),  # ARABIC LETTER ZAIN
        "\u0633": ("s", "s", "s", "s"),  # ARABIC LETTER SEEN
        "\u0634": ("$", "^s", "š", "sh"),  # ARABIC LETTER SHEEN
        "\u0635": ("S", "*s", "ṣ", "ṣ"),  # ARABIC LETTER SAD
        "\u0636": ("D", "*d", "ḍ", "ḍ"),  # ARABIC LETTER DAD
        "\u0637": ("T", "*t", "ṭ", "ṭ"),  # ARABIC LETTER TAH
        "\u0638": ("Z", "*z", "ẓ", "ẓ"),  # ARABIC LETTER ZAH
        "\u0639": ("E", "`", "ʿ", "`"),  # ARABIC LETTER AIN
        "\u063a": ("g", "*g", "ġ", "gh"),  # ARABIC LETTER GHAIN
        "\u0640": ("_", "", "", ""),  # ARABIC TATWEEL
        "\u0641": ("f", "f", "f", "f"),  # ARABIC LETTER FEH
        "\u0642": ("q", "*k", "ḳ", "q"),  # ARABIC LETTER QAF
        "\u0643": ("k", "k", "k", "k"),  # ARABIC LETTER KAF
        "\u0644": ("l", "l", "l", "l"),  # ARABIC LETTER LAM
        "\u0645": ("m", "m", "m", "m"),  # ARABIC LETTER MEEM
        "\u0646": ("n", "n", "n", "n"),  # ARABIC LETTER NOON
        "\u0647": ("h", "h", "h", "h"),  # ARABIC LETTER HEH
        "\u0648": ("w", "w", "w", "w"),  # ARABIC LETTER WAW
        "\u0649": ("Y", "/a", "á", "ā"),  # ARABIC LETTER ALEF MAKSURA
        "\u064a": ("y", "y", "y", "y"),  # ARABIC LETTER YEH
        "\u064b": ("F", "a*n", "aⁿ", "an"),  # ARABIC FATHATAN
        "\u064c": ("N", "u*n", "uⁿ", "un"),  # ARABIC DAMMATAN
        "\u064d": ("K", "i*n", "iⁿ", "in"),  # ARABIC KASRATAN
        "\u064e": ("a", "a", "a", "a"),  # ARABIC FATHA
        "\u064f": ("u", "u", "u", "u"),  # ARABIC DAMMA
        "\u0650": ("i", "i", "i", "i"),  # ARABIC KASRA
        "\u0651": ("~", "u", "u", "ūw"),  # ARABIC SHADDA
        "\u0652": ("o", "a", "a", "a"),  # ARABIC SUKUN
        "\u0653": ("^", "_a", "ā", "ā"),  # ARABIC MADDAH ABOVE
        "\u0654": ("#", "'", "ʾ", "ā"),  # ARABIC HAMZA ABOVE
        "\u0655": ("=", "'", "ʾ", "ā"),  # ARABIC HAMZA BELOW
        "\u0660": ("0", "0", "0", "0"),  # ARABIC INDIC DIGIT ZERO
        "\u0661": ("1", "1", "1", "1"),  # ARABIC INDIC DIGIT ONE
        "\u0662": ("2", "2", "2", "2"),  # ARABIC INDIC DIGIT TWO
        "\u0663": ("3", "3", "3", "3"),  # ARABIC INDIC DIGIT THREE
        "\u0664": ("4", "4", "4", "4"),  # ARABIC INDIC DIGIT FOUR
        "\u0665": ("5", "5", "5", "5"),  # ARABIC INDIC DIGIT FIVE
        "\u0666": ("6", "6", "6", "6"),  # ARABIC INDIC DIGIT SIX
        "\u0667": ("7", "7", "7", "7"),  # ARABIC INDIC DIGIT SEVEN
        "\u0668": ("8", "8", "8", "8"),  # ARABIC INDIC DIGIT EIGHT
        "\u0669": ("9", "9", "9", "9"),  # ARABIC INDIC DIGIT NINE
        "\u0670": ("`", "~a", "ã", ""),  # ARABIC LETTER SUPERSCRIPT ALEF
        "\u0671": ("{", "a", "a", "a"),  # ARABIC LETTER ALEF WASLA
        "\u06af": ("G", "g", "g", "g"),  # ARABIC LETTER GAF
        "\u06cc": ("J", "#y", "Y", "y"),  # ARABIC LETTER FARSI YEH
        "\u06d6": (
            "SlY",
            "*sl/a",
            "ṣlá",
            "ṣla",
        ),  # ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA
        "\u06da": ("M", "#j", "J", "j"),  # ARABIC SMALL HIGH JEEM
        "\u06dc": (":", "#s", "S", "s"),  # ARABIC SMALL HIGH SEEN
        "\u06df": ("@", "0", "0", "0"),  # ARABIC SMALL HIGH ROUNDED ZERO
        "\u06e0": ('"', "0", "0", "0"),  # ARABIC SMALL HIGH UPRIGHT RECTANGULAR ZERO
        "\u06e2": ("[", "#m", "M", "M"),  # ARABIC SMALL HIGH MEEM ISOLATED FORM
        "\u06e3": (";", "#s", "S", "S"),  # ARABIC SMALL LOW SEEN
        "\u06e5": (",", "#w", "W", "W"),  # ARABIC SMALL WAW
        "\u06e6": (".", "#y", "Y", "Y"),  # ARABIC SMALL YEH
        "\u06e8": ("!", "#n", "N", "N"),  # ARABIC SMALL HIGH NOON
        "\u06ea": ("-", ".", ".", "."),  # ARABIC EMPTY CENTRE LOW STOP
        "\u06eb": ("+", ".", ".", "."),  # ARABIC EMPTY CENTRE HIGH STOP
        "\u06ec": ("%", ".", ".", "."),  # ARABIC ROUNDED HIGH STOP WITH FILLED CENTRE
        "\u06ed": ("]", "#m", "M", "M"),  # ARABIC SMALL LOW MEEM
    }
    """
    More Arabic transcriptions:

    *   column 1: `custom` [Quran-tanzil](http://tanzil.net/#1:1), slightly extended
    *   column 2: `ascii` resp. `latin` plus diacritics also known as betacode.
        We use a list compiled by
        [Peter Verkinderen](https://pverkind.github.io/betacodeTranscriber/js/betacode.js)
    *   column 4: `standard` (Library of Congress) (to-be filled).
        We use the
        [arabic romanization list of 2012](https://www.loc.gov/catdir/cpso/romanization/arabic.pdf)
        We refrain of from applying rules that cannot be computed without
        lexical/grammatical/dialectical knowledge of the arabic language.
    """

    arabicTransQuran = {}
    arabicTransAscii = {}
    arabicTransLatin = {}
    arabicTransStandard = {}

    for ara, (qur, asc, lat, std) in arabicTrans.items():
        arabicTransQuran[ara] = qur
        arabicTransAscii[ara] = asc
        arabicTransLatin[ara] = lat
        arabicTransStandard[ara] = std

    def quranFromArabic(word):
        return "".join(Transcription.arabicTransQuran.get(x, x) for x in word)

    def asciiFromArabic(word):
        return "".join(Transcription.arabicTransAscii.get(x, x) for x in word)

    def latinFromArabic(word):
        return "".join(Transcription.arabicTransLatin.get(x, x) for x in word)

    def standardFromArabic(word):
        return "".join(Transcription.arabicTransStandard.get(x, x) for x in word)

    def __init__(self):
        self.hebrew_consonants = {
            Transcription.hebrew_mapping[x] for x in Transcription.hebrew_cons
        }
        self.hebrew_consonants.add("\u05E9")
        self.hebrew_mappingi = dict(
            (v, k) for (k, v) in Transcription.hebrew_mapping.items() if k != ""
        )
        self.syriac_mappingi = dict(
            (v, k) for (k, v) in Transcription.syriac_mapping.items() if k != ""
        )
        # special treatment needed for nun hafukha,
        # since it is consists of two characters
        self.hebrew_mappingi["\u05C6"] = "ñ"
        self.hebrew_mappingi["\u0307"] = ""
        self.syriac_punctuation_trans = (
            "#\\",
            "=.",
            "=#",
            "=:",
            "=^",
            "=/",
            "=\\",
            "^:",
            "^\\",
        )
        self.syriac_punctuation_syc = tuple(
            Transcription.syriac_mapping[c] for c in self.syriac_punctuation_trans
        )

        arabicMap = dict(qur={}, asc={}, lat={}, std={})
        arabicMapI = dict(qur={}, asc={}, lat={}, std={})

        for u in sorted(Transcription.arabicTrans):
            (qur, asc, lat, std) = Transcription.arabicTrans[u]
            for k, v in zip(arabicMap.keys(), Transcription.arabicTrans[u]):
                arabicMap[k].setdefault(v, u)
                arabicMapI[k][u] = v

        self.arabicMap = arabicMap
        self.arabicMapI = arabicMapI

    def sycSplitPunc(self):
        pass

    def _comp(s):
        for d, c in Transcription.decomp.items():
            s = s.replace(d, c)
        return s

    def _decomp(s):
        for d, c in Transcription.decomp.items():
            s = s.replace(c, d)
        return s

    def suffix_and_finales(word):
        """
        Given an ETCBC transliteration, split it into the word material
        and the interword material that follows it (space, punctuation).
        Replace the last consonant of the word material by its final form, if applicable.

        Output a tuple with the modified word material and the interword material.

        Example:

        ``` python
        Transcription.suffix_and_finales('71T_H@>@95REY00')
        ```

        Output:

        ```
        ('71T_H@>@95REy', '00\n')
        ```

        Note that the `Y` has been replaced by `y`.
        """

        # first split the word proper from the suffix,
        # and add a space if there is no other suffix
        add_space = ""
        suffix = ""
        new_word = word
        if not word:
            return (new_word, suffix + add_space)
        lastch = new_word[-1]
        if lastch == "-" or lastch == "&":
            new_word = new_word[0:-1]
            suffix = lastch
        else:
            if len(new_word) >= 2:
                lastch = new_word[-1]
                llastch = new_word[-2]
                if llastch == "_" and (lastch == "P" or lastch == "S"):
                    new_word = new_word[0:-2]
                    suffix = " " + lastch + suffix + " "
            if len(new_word) >= 2:
                lastch = new_word[-1]
                llastch = new_word[-2]
                if llastch == "_" and (lastch == "N"):
                    new_word = new_word[0:-2]
                    suffix = " ñ" + suffix + " "
            if len(new_word) >= 2:
                lastch = new_word[-1]
                llastch = new_word[-2]
                if llastch == "0" and (lastch == "0" or lastch == "5"):
                    new_word = new_word[0:-2]
                    suffix = (" " if lastch == "5" else "") + llastch + lastch + suffix
                    add_space = "\n" if lastch == "0" else " "
        if suffix == "":
            add_space = " "
        elif suffix == "-":
            add_space = ""
            suffix = ""
        # second: replace consonants by their final forms when needed
        new_word = Transcription.trans_final_pat.sub(Transcription._map_final, new_word)
        return (new_word, suffix + add_space)

    def _map_final(m):
        return m.group(1) + m.group(2).lower() + m.group(3)

    def _map_hebrew(m):
        return Transcription.hebrew_mapping.get(m.group(1), m.group(1))

    def _map_syriac(m):
        return Transcription.syriac_mapping.get(m.group(1), m.group(1))

    def _swap_accent(m):
        return m.group(1) + m.group(3) + m.group(4) + m.group(2)

    def _remove_accent(m):
        return "00" if m.group(1) == "00" else "05" if m.group(1) == "05" else ""

    def _remove_point(m):
        return "00" if m.group(1) == "00" else "05" if m.group(1) == "05" else ""

    def _ph_simple(m):
        return "å" if m.group(1) in "āo" else ""

    # UNICODE normalization is harmful
    # if there is a combination of dagesh, vowel and accent.

    def suppress_space(word):
        """
        Given an ETCBC transliteration of a word,
        match the end of the word for punctuation and spacing characters
        (`sof pasuq`, `paseq`, `nun hafukha`, `setumah`, `petuhah`, space, no-space)

        Example:

        ``` python
        Transcription.suppress_space('B.:&')
        Transcription.suppress_space('B.@R@74>')
        Transcription.suppress_space('71T_H@>@95REY00')
        ```

        Output:

        ```
        <re.Match object; span=(3, 4), match='&'>
        None
        <re.Match object; span=(13, 15), match='00'>
        ```
        """

        return Transcription.noorigspace.search(word)

    def to_etcbc_v(word):
        """
        Given an ETCBC transliteration of a fully pointed word,
        strip all the non-vowel pointing (i.e. the accents).

        Example:

        ``` python
        Transcription.to_etcbc_v('HAC.@MA73JIm')
        ```

        Output:

        ```
        HAC.@MAJIm
        ```
        """

        return Transcription.remove_accent_pat.sub(Transcription._remove_accent, word)

    def to_etcbc_c(word):
        """
        Given an ETCBC transliteration of a fully pointed word,
        strip everything except the consonants.
        Punctuation will also be stripped.

        Example:

        ``` python
        Transcription.to_etcbc_c('HAC.@MA73JIm')
        ```

        Output:

        ```
        H#MJM
        ```

        Note that the pointed shin (`C`) is replaced by an unpointed one (`#`).
        """

        word = Transcription.remove_point_pat.sub(Transcription._remove_point, word)
        word = Transcription.remove_psn_pat.sub(
            "00", word
        )  # remove nun hafukha, setumah, petuhah at the end of a verse
        word = Transcription.remove_psq_pat.sub(
            " ", word
        )  # replace paseq with attached spaces by single space
        word = word.upper()  # no final forms of consonants
        return Transcription.shin_pat.sub("#", word)

    def to_hebrew(word):
        """
        Given a transliteration of a fully pointed word,
        produce the word in UNICODE Hebrew.
        Care will be taken that vowel pointing will be added to consonants
        before accent pointing.

        Example:

        ``` python
        Transcription.to_hebrew('HAC.@MA73JIm')
        ```

        Output:

        ```
        הַשָּׁמַ֖יִם
        ```
        """

        word = Transcription.swap_accent_pat.sub(Transcription._swap_accent, word)
        return Transcription.trans_hebrew_pat.sub(Transcription._map_hebrew, word)

    def to_hebrew_v(word):
        """
        Given a transliteration of a fully pointed word,
        produce the word in UNICODE Hebrew, but without the accents.

        Example:

        ``` python
        Transcription.to_hebrew_v('HAC.@MA73JIm')
        ```

        Output:

        ```
        הַשָּׁמַיִם
        ```
        """

        return Transcription.trans_hebrew_pat.sub(
            Transcription._map_hebrew, Transcription.to_etcbc_v(word)
        )

    def to_hebrew_c(word):
        """
        Given a transliteration of a fully pointed word,
        produce the word in UNICODE Hebrew, but without the pointing.

        Example:

        ``` python
        Transcription.to_hebrew_c('HAC.@MA73JIm')
        ```

        Output:

        ```
        השמימ
        ```

        Note that final consonant forms are not being used.
        """

        return Transcription.trans_hebrew_pat.sub(
            Transcription._map_hebrew, Transcription.to_etcbc_c(word)
        )

    def to_hebrew_x(word):
        """
        Given a transliteration of a fully pointed word,
        produce the word in UNICODE Hebrew, but without the pointing.
        Vowel pointing and accent pointing will be applied in the order given
        by the input word.

        Example:

        ``` python
        Transcription.to_hebrew_x('HAC.@MA73JIm')
        ```

        Output:

        ```
        הַשָּׁמַ֖יִם
        ```
        """

        return Transcription.trans_hebrew_pat.sub(Transcription._map_hebrew, word)

    def ph_simplify(pword):
        """
        Given a phonological transliteration of a fully pointed word,
        produce a more coarse phonological transliteration.

        Example:

        ``` python
        Transcription.ph_simplify('ʔᵉlōhˈîm')
        Transcription.ph_simplify('māqˈôm')
        Transcription.ph_simplify('kol')
        ```

        Output:

        ```
        ʔlōhîm
        måqôm
        kål
        ```

        Note that the simplified version transliterates the `qamets gadol` and `qatan`
        to the same character.
        """

        return Transcription.ph_simple_pat.sub(Transcription._ph_simple, pword)

    def from_hebrew(self, word):
        """
        Given a fully pointed word in UNICODE Hebrew,
        produce the word in ETCBC transliteration.

        Example:

        ``` python
        tr.from_hebrew('הָאָֽרֶץ׃')
        ```

        Output:

        ```
        H@>@95REy00
        ```
        """

        return "".join(
            self.hebrew_mappingi.get(x, x) for x in Transcription._comp(word)
        )

    def to_syriac(self, word):
        """
        Given a word in ETCBC transliteration,
        produce the word in UNICODE Syriac.

        Example:

        ``` python
        tr.to_syriac('MKSJN')
        ```

        Output:

        ```
        ܡܟܣܝܢ
        ```
        """

        return Transcription.trans_syriac_pat.sub(Transcription._map_syriac, word)

    def from_syriac(self, word):
        """
        Given a word in UNICODE Syriac,
        produce the word in ETCBC transliteration.

        Example:

        ``` python
        tr.from_syriac('ܡܟܣܝܢ')
        ```

        Output:

        ```
        MKSJN
        ```
        """

        return "".join(self.syriac_mappingi.get(x, x) for x in word)

    def can_to_syriac(self, word):
        return all(
            candidate in Transcription.syriac_mapping
            for candidate in Transcription.trans_syriac_pat.findall(word)
            if candidate != " "
        )

    def can_from_syriac(self, word):
        return all(c in self.syriac_mappingi for c in word if c != " ")

    def to_ugaritic(word):
        """
        Given a word in transliteration,
        produce the word in UNICODE Ugaritic.

        kṯbx
        𐎋𐎘𐎁x


        Example:

        ``` python
        Transcription.to_ugaritic('kṯbx')
        ```

        Output:

        ```
        𐎋𐎘𐎁x
        ```
        """

        return "".join(Transcription.ugaritic_mapping.get(x, x) for x in word)

    def from_ugaritic(word):
        """
        Given a word in UNICODE Ugaritic,
        produce the word in transliteration.

        Example:

        ``` python
        Transcription.from_ugaritic('𐎋𐎘𐎁x')
        ```

        Output:

        ```
        kṯbx
        ```
        """

        return "".join(Transcription.ugaritic_mappingi.get(x, x) for x in word)

    def to_arabic(word):
        """
        Given a word in transliteration,
        produce the word in UNICODE Arabic.

        Example:

        ``` python
        Transcription.to_arabic('bisomi')
        ```

        Output:

        ```
        بِسْمِ
        ```
        """

        return "".join(Transcription.arabic_mapping.get(x, x) for x in word)

    def from_arabic(word):
        """
        Given a word in UNICODE Arabic,
        produce the word in transliteration.

        Example:

        ``` python
        Transcription.from_arabic('بِسْمِ')
        ```

        Output:

        ```
        bisomi
        ```
        """

        return "".join(Transcription.arabic_mappingi.get(x, x) for x in word)



"""
Enable manual annotation APIs.

Currently only `tf.ner.ner` is supported here,
but other annotation tools might be added in the future.
"""

import types

from ..ner.ner import NER


def makeNer(app, silent=False, **kwargs):
    """Produce an instance of the NER class."""
    return NER(app, silent=silent, **kwargs)


def annotateApi(app):
    """Produce the interchange functions API.

    Parameters
    ----------
    app: obj
        The high-level API object
    """

    app.makeNer = types.MethodType(makeNer, app)



"""
# Access to `oslots` feature data.

In general, features are stored as dictionaries, but this specific feature
has an optimised representation. Since it is a large feature and present
in any TF dataset, this pays off.
"""


class OslotsFeature:
    def __init__(self, api, metaData, data):
        self.api = api
        self.meta = metaData
        """Metadata of the feature.

        This is the information found in the lines starting with `@`
        in the `.tf` feature file.
        """

        self.data = data[0]
        self.maxSlot = data[1]
        self.maxNode = data[2]

    def items(self):
        """A generator that yields the non-slot nodes with their slots.
        """

        maxSlot = self.maxSlot
        data = self.data
        maxNode = self.maxNode

        shift = maxSlot + 1

        for n in range(maxSlot + 1, maxNode + 1):
            yield (n, data[n - shift])

    def s(self, n):
        """Get the slots of a (non-slot) node.

        Parameters
        ----------
        node: integer
            The node whose slots must be retrieved.

        Returns
        -------
        tuple
            The slot nodes of the node in question, in canonical order.
            (`tf.core.nodes`)

            For slot nodes `n` it is the tuple `(n,)`.

            All non-slot nodes are linked to at least one slot.
        """

        if n == 0:
            return ()
        if n < self.maxSlot + 1:
            return (n,)
        m = n - self.maxSlot
        if m <= len(self.data):
            return self.data[m - 1]
        return ()



"""
# Parameters

Fixed values for the whole program.
"""

import os
import sys
from zipfile import ZIP_DEFLATED


VERSION = '13.1.0'
"""Program version.

This value is under control of the update process, as run by
`build.py` in the top-level directory of the repo.
"""

NAME = "Text-Fabric"
"""The name of the game: this program.
"""

BANNER = f"This is {NAME} {VERSION}"

PACK_VERSION = "4"
"""Data serialization version.

Plain text feature files will be compressed to zipped, `pickled` data structures
that load must faster.

These methods evolve, sometimes in incompatible ways.
In those cases we bump this version number.
That will cause TF not to use compressed files that have been compressed by
older, incompatible methods.
Instead, TF will produce freshly compressed data files.

The compressed data files are stored in a directory `.tf/{PVN}/` next
to the original `tf`  files, where `{PVN}` is the package version number.

See Also
--------
tf.clean
"""

API_VERSION = 3
"""TF API version.

TF offers and API to TF apps.
This is the version that the current TF offers to its apps.

Apps require a version. The provided version and the required version must
match exactly in order to get a working system.

We do not aim for backward compatibility, since it is very easy to obtain a new version
of an app.

When TF loads a TF app, it will check the api version that the app requires
against this version.

App requirement higher than TF API version
:   The user is advised to upgrade TF, or, alternatively,
    select an older version of the app

App requirement lower than TF API version
:   The user is advised to obtain a newer version of the app, or alternatively,
    downgrade TF
"""

OTYPE = "otype"
"""Name of a central feature in a TF data set:
`otype` maps nodes to their types."""

OSLOTS = "oslots"
"""Name of a central feature in a TF data set:
`oslots` maps non-slot nodes to the sets of slots they occupy."""

OTEXT = "otext"
"""Name of a central (but optional) feature in a TF data set:
`otext` has configuration settings for sections, structure, and text formats."""

OVOLUME = "ovolume"
"""Name of the feature that maps nodes of a work dataset
to nodes in individual volumes in that work."""

OWORK = "owork"
"""Name of the feature that maps nodes in an individual volume of a work
to nodes in that work."""

OINTERF = "ointerfrom"
"""Name of the feature that stores the outgoing inter-volume edges
of a volume."""

OINTERT = "ointerto"
"""Name of the feature that stores the incoming inter-volume edges
of a volume."""

OMAP = "omap"
"""Name prefix of features with a node map from an older version to a newer version.

The full name of such a feature is `omap@`*oldversion*`-`*newversion*
"""

WARP = (OTYPE, OSLOTS, OTEXT)
"""The names of the central features of TF datasets.

The features `otype` and `oslots` are crucial to every TF dataset.
Without them, a dataset is not a TF dataset, although it could still be a
TF data module.
"""

GZIP_LEVEL = 2
"""Compression level when compressing TF files."""

PICKLE_PROTOCOL = 4
"""Pickle protocol level when pickling TF files."""

ORG = "annotation"
"""GitHub organization or GitLab group.

This is where the repo that contains TF resides.
"""
REPO = "text-fabric"
"""GitHub repo or GitLab project.

This is the name of the repo that contains TF.
"""

RELATIVE = "tf"
"""Default relative path with a repo to the directory with TF files.
"""

ON_IPAD = sys.platform == "darwin" and os.uname().machine.startswith("iP")

GH = "github"
"""Name of GitHub backend."""

GL = "gitlab"
"""Name of GitLab backend."""

URL_GH = "https://github.com"
"""Base URL of GitHub."""

URL_GH_API = "https://api.github.com"
"""Base URL of GitHub API."""

URL_GH_UPLOAD = "https://uploads.github.com"
"""Base URL of GitHub upload end point."""

URL_GL = "https://gitlab.com"
"""Base URL of GitLab."""

URL_GL_API = "https://api.gitlab.com"
"""Base URL of GitLab API."""

URL_GL_UPLOAD = "https://uploads.gitlab.com"
"""Base URL of GitLab upload end point."""

URL_NB = "https://nbviewer.jupyter.org"
"""Base URL of NB-viewer."""

URL_TF_DOCS = "https://annotation.github.io/text-fabric/tf"


PROTOCOL = "http://"
HOST = "localhost"  #
PORT_BASE = 10000


DOI_DEFAULT = "no DOI"
DOI_URL_PREFIX = "https://doi.org"

DOI_TF = "10.5281/zenodo.592193"
"""DOI of an archived copy of this repo at Zenodo."""

BRANCH_DEFAULT = "master"
"""Default branch in repositories, older value."""

BRANCH_DEFAULT_NEW = "main"
"""Default branch in repositories, modern value."""

ZIP_OPTIONS = dict(compression=ZIP_DEFLATED)
"""Options for zip when packing TF files.

This is for packaging collections of plain TF files into zip files
to be attached to releases on GitHub / GitLab.

!!! caution "Not for `.tfx` files"
    This is not the zipping as done when `.tf` files are
    `pickled` and compressed to `.tfx` files.
"""

if sys.version_info[0] > 3 or sys.version_info[0] == 3 and sys.version_info[1] >= 7:
    ZIP_OPTIONS["compresslevel"] = 6

YARN_RATIO = 1.25
"""Performance parameter in the `tf.search.search` module."""

TRY_LIMIT_FROM = 40
"""Performance parameter in the `tf.search.search` module."""

TRY_LIMIT_TO = 40
"""Performance parameter in the `tf.search.search` module."""

SEARCH_FAIL_FACTOR = 4
"""Limits fetching of search results to this times maxNode (corpus dependent)"""
